 /****************************************************************
 *
 *
 *   a3_Plane.cpp
 *
 *   An arrangement of planes, as a class.
 *   Used for ArrangePak-3D.
 *
 *
 *   ArrangePak-3D 
 *   version 1.04
 *   original version:   October, 2003
 *   this revision:      May 2004
 *
 *
 *   This is the class library for an arrangement of planes in 3-dimensional
 *   space. An arrangement of planes maintains a list of planes in 3-space,
 *   a 'viewLine' attribute that represents the intersection of the given
 *   plane with a viewport at 'z'=0 in the cartesian coordinate system,   
 *   and additional member attributes such as 'totalPlanes', the number of
 *   planes in the arrangement, as well as necessary member functions.
 *
 *------------------------------------------------------------------------
 *
 *   copyright (c) 2004, 2003 Stephen Wismath, Ph.D.
 *   University of Lethbridge
 *   Lethbridge, AB   CANADA
 *
 *------------------------------------------------------------------------
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *------------------------------------------------------------------------
 *
 *   Licence terms do not apply to the Library of Efficient Data Types and 
 *   Algorithms, which is licensed separately by 
 *       Algorithmic Solutions Software Gmbh 
 *       http://www.algorithmic-solutions.com/
 *
 *************************************************************************/
 
#include <LEDA/random.h>                  // random generator, for random points
#include <LEDA/circle.h>                    // for testing whether planes bisect the viewport
#include <LEDA/vector.h>
#include <LEDA/d3_line.h>
#include <LEDA/graph_iterator.h>
#include <LEDA/graphwin.h>
#include <LEDA/file.h>

#include <iostream>

#include "a3_Plane.h"
#include "graphNode.h"
#include "parameterNode.h"

extern leda_point originPt;                  // the origin for our 2D window
extern double mainCircleRadius;         // the radius of our viewport
extern double xOffset;
extern double yOffset;
extern double zVal;
extern const leda_d3_plane frontFace;   // imaginary plane between the display and the user
extern leda_string filePath;                    // working directory


/********************************************************************
 *
 * constructor
 *
 * when object is created, LEDA automatically constructs an
 *  empty list of 'aPlane' objects for it
 *
 *******************************************************************/
a3_Plane::a3_Plane()
{
  hasChanged = true;
  totalPlanes = 0;
}

/******************************************************************
 *
 *  destructor does nothing, LEDA memory manager handles the list
 *
 *****************************************************************/
a3_Plane::~a3_Plane() 
{ }


/**************************************************************
 *  function adds a DEFAULT plane to the arrangement
 *
 *  returns updated total number of planes in the arrangement 
 *
 *  pre: takes leda_window by reference, must exist!
 *       leda_window is not modified, but required for dimensions
 * 
 *       number of planes in arrangement must be less than
 *        maxPlaneCount
 *
 *  post: returns updated total number of planes in arrangement
 *        if arrangement already has maximum number of planes
 *         then function returns -1        
 *  
 **************************************************************/
int a3_Plane::add_Default_Plane()
{
  if (isFull())          // not permitted to add more planes
     return -1;          // so exit the function with an error

  // start with a unit vector, normal to the default (horizontal) plane
  // adjusted from the LEDA origin to our universe origin
  leda_d3_point p1(originPt.xcoord(), originPt.ycoord(), 0);
  leda_vector v1(0.0, 1.0, 0.0);
  leda_d3_plane defPlane(p1, v1);
  leda_d3_point q, r;
  leda_list<leda_point> pointList;
  
  if (intersectsFace(defPlane, q, r ) == 1)
    {
      leda_d3_line line1(q,r);
      
      leda_line line2;                             // line2 will be the 2D line of intersection between plane and circle
      if (line1.project_xy(line2)) {
              // now test whether the 2D line intersects the viewport circle
	// if true, the plane is valid for use in our arrangement
	// otherwise, the plane is outside our universe, so reject it
	leda_circle circ(originPt, mainCircleRadius);
	pointList = circ.intersection(line2);
	
	if (pointList.size() == 2) {
	  // add plane to list of planes, include corresponding line, and increment the count
	
	  leda_line newLine(pointList.front(), pointList.back());
	  aPlane newPlane(defPlane, newLine, ++totalPlanes);
	  planeList.append(newPlane);
	  hasChanged = true;
	}
      }
    }
  return totalPlanes;        // return new total number of planeItems in vector
}



/**************************************************************
 *  function adds a RANDOM plane to the arrangement
 *
 *  returns updated total number of planes in the arrangement 
 *
 *  pre: takes leda_window by reference, must exist!
 *       leda_window is not modified, but required for dimensions
 * 
 *       number of planes in arrangement must be less than
 *        maxPlaneCount
 *
 *  post: returns updated total number of planes in arrangement
 *        if arrangement already has maximum number of planes
 *         then function returns -1        
 *  
 **************************************************************/
int a3_Plane::add_Random_Plane()
{
   if (isFull())               // not permitted to add more planes
     return -1;                // so exit the function with an error

   leda_d3_point q, r;
   leda_list<leda_point> pointList;
   
   while (true) {      // remain in loop until a valid plane is generated
     
     // make a random plane from three random points in 3D space
     leda_d3_plane randPlane(random_3d_point(),
			     random_3d_point(),
			     random_3d_point());
     
     if (intersectsFace(randPlane, q, r ) == 1 ) {
       leda_d3_line line1(q,r);
       
       leda_line line2;                // line2 will be the 2D line of intersection between plane and circle
       if (line1.project_xy(line2)) {

	  //    now test whether the 2D line intersects the viewport circle
	  //    if true, the plane is valid for use in our arrangement
	  //    otherwise, the plane is outside our universe, so reject it
	  
	 leda_circle circ(originPt, mainCircleRadius);
	 pointList = circ.intersection(line2);
	 
	 if (pointList.size() == 2) {
	   // add plane to list of planes, include corresponding line, and increment the count
	   leda_line newLine(pointList.front(), pointList.back());
	   aPlane newPlane(randPlane, newLine, ++totalPlanes);
	   planeList.append(newPlane);
	   hasChanged = true;
	   return totalPlanes;        // return new total number of planeItems in vector
	 }
       }
     }
   }
   return -1;       // should NEVER see this
}



/**************************************************************
 *  function removes a 3-D plane from the arrangement
 *
 *  pre: takes x and y coordinates of plane to delete
 *      
 *  post: returns updated number of planes in the arrangement
 *
 *  finds and deletes whichever plane in the vector is closest
 *   to the (x,y) coordinates which are the input arguments
 *   
 *  IF MORE THAN ONE LINE IS THE SAME DISTANCE TO THE POINT
 *  THE ONE DELETED IS THE THE ONE WITH LEAST INDEX IN THE VECTOR
 *
 **************************************************************/
int a3_Plane::delete_Plane(double x, double y)
{
  list_item it;                    // temporary iterator
  list_item itemToDelete;
  double distance;
  leda_point clickPt(x, y);
  
  if (! planeList.empty()) {
    // to begin, we mark the first plane in the list for deletion
    distance = (planeList.front().getLine()).distance(clickPt);
    itemToDelete = planeList.first();
    
    // find a plane closer to the mouse click, and delete that one instead
    forall_items(it, planeList) {
      if (distance > ((planeList.contents(it)).getLine()).distance(clickPt)) {
	distance = ((planeList.contents(it)).getLine()).distance(clickPt);
	itemToDelete = it;
      }
    } 
    planeList.erase(itemToDelete);
    totalPlanes--;                       // decrement total count
    hasChanged = true;
  }
  return planeList.size();
}



/***************************************************************
 *  function replaces an 'aPlane' item in the arrangement list
 *    with a new 'aPlane' item containing different values
 *
 **************************************************************/
void a3_Plane::swapPlane(leda_d3_plane& planeIn, int planeNum)
{
  leda_d3_point q, r;
  leda_list<leda_point> pointList;
  list_item it;
  
  if (intersectsFace(planeIn, q, r ) == 1) {
    leda_d3_line line1(q,r);
    
    leda_line line2;                // line2 will be the 2D line of intersection between plane and circle
    if (line1.project_xy(line2)) {
       /*
	 now test whether the 2D line intersects the viewport circle
	 if true, the plane is valid for use in our arrangement
	 otherwise, the plane is outside our universe, so reject it
       */
      leda_circle circ(originPt, mainCircleRadius);
      pointList = circ.intersection(line2);
      
      if (pointList.size() == 2) {
	// swap the plane
	leda_line newLine(pointList.front(), pointList.back());
	it = planeList.get_item(planeNum-1);
	aPlane newPlane(planeIn, newLine, planeList.inf(it).getIndex());
	planeList.assign(it, newPlane);
      }
    }
  }
  return;
}
  

  

/***************************************************************
 *
 * Function returns a 2-Dimensional line identical to 
 *  the XY-projection of the corresponding plane onto Z=0.
 *
 * The plane is selected from a list of a3_Plane data types.
 *
 * pre: planeList must not be empty
 *      planeList is zero-indexed, but input argument is NOT
 *      e.g. if there are 10 planes in the list, 
 *
 *           1 <= (input arg) <= 10
 *
 * post: returns a line corresponding to the indexed plane             
 *
 ***************************************************************/
leda_line a3_Plane::get_2D_Line(int i)
{
  
  leda_assert(planeList.size() >= i,
	      "a3_Plane::getLine - past end of list", -1);

  return(planeList.contents(planeList.get_item(i-1))).getLine();
 
}



/***************************************************************
 *
 *  returns a segment in 3D space corresponding to that portion
 *   of a plane which intersects the viewport at (x,y,0)
 *
 *  the 3D plane under test is given by an index into the list 
 *
 *
 **************************************************************/
leda_d3_segment a3_Plane::get_XY_Axis(int planeNum)
{
  leda_d3_point q, r;
  leda_d3_plane tempPlane;
  getPlane(planeNum, tempPlane);
  intersectsFace(tempPlane, q, r);
  leda_d3_segment result(q,r);
  return result;
}




/***************************************************************
 *
 * returns a segment in 3D space defined by a point in Z+ 
 *  to its reflected point in Z-
 *
 * the 3D segment can be used as a pviot axis around
 *  which to rotate the plane
 *
 **************************************************************/
leda_d3_segment a3_Plane::get_3D_Axis(int planeNum, const leda_point &p1, 
				      const leda_point& p2, int pointNum)
{
  leda_d3_point q(p1.xcoord()+xOffset, p1.ycoord()+yOffset, zVal);
  leda_d3_point r(p2.xcoord()+xOffset, p2.ycoord()+yOffset, zVal);
  
  leda_d3_point orgPt(0.0, 0.0, 0.0);
  leda_d3_plane tempPlane;
  
  if (pointNum = -1)
    r = q.reflect(orgPt);
  else
    q = r.reflect(orgPt);
  
  leda_d3_segment result(q,r);
  return result;
}



/************************************************************
 * 
 *  returns a point corresponding along the intersection 
 *   line between a given plane and the frontFace plane
 *
 *  takes an index into a list of planes, to specify which plane
 *
 *  use pointNum to specify which point we want the function
 *   to return
 *
 *  valid input values for pointNum = { 0, +1, -1 }
 *
 *  if pointNum is not valid, default to pointNum = 0 and return
 *   a midpoint between q and r
 *
 ***********************************************************/
leda_d3_point a3_Plane::getRotationPoint(int planeNum, int pointNum)
{
  leda_d3_point q, r;
  leda_d3_plane tempPlane;

  getPlane(planeNum, tempPlane);

  frontFace.intersection(tempPlane, q, r);       // frontFace is globally defined

  if (pointNum == 1)
    return q;
  
  if (pointNum == -1)
    return r;
  
  return midpoint(q,r);

}




/******************************************************************
 *
 *  get plane from a list of objects
 *  takes an integer argument as index into the list
 *  returns 3D plane by reference
 *
 *
 *  returns 0 if success, otherwise returns non-zero (1)
 *
 *****************************************************************/
int a3_Plane::getPlane(int planeNum, leda_d3_plane &plResult)
{
  list_item it;
  if (planeList.empty() || planeList.size() < planeNum)
    return 1;
  
  plResult = (planeList.contents(planeList.get_item(planeNum-1))).getPlane();
  return 0;
} 



/********************************************************************
 *
 *  locate an item in the list of planes
 *  return the index to the corresponding entry in the list
 *
 *  pre: list must contain at least one item
 *       cannot search through an empty list of planes!
 *******************************************************************/
int a3_Plane::findPlane(double x, double y)
{
   list_item it;                    // temporary iterators
   list_item itemToFind;
  
   double distance;
   leda_point clickPt(x, y);
   
   // to begin, we "find" the first plane in the list
   distance = (planeList.front().getLine()).distance(clickPt);
   itemToFind = planeList.first();
   
   // if there is a plane closer to the mouse click, "find" that one instead
   forall_items(it, planeList) {
     if (distance > ((planeList.contents(it)).getLine()).distance(clickPt))	 {
       distance = ((planeList.contents(it)).getLine()).distance(clickPt);
       itemToFind = it;
     }
   }
   int result =  planeList.contents(itemToFind).getIndex();
   return result;
}
   




/*************************************************************************
*
*  driver function generates an undirected graph from an arrangement
*   of planes in 3D space
*
*  returns a numeric status code to the calling function
*
*  also returns 3 planes by reference: if there is an error generating 
*    the graph, it was caused while attempting to find a point of intersection
*    for the three given planes
*
*  if the graph succeeds, the three planes are empty references when this
*   function returns
*    
**************************************************************************/
int a3_Plane::to_Graph(const int outputType, const leda_string &outputFile,
		       int &plane1, int &plane2, int &plane3)
{
  plane1 = 0;
  plane2 = 0;
  plane3 = 0;
  return extract_Graph(outputType, outputFile, plane1, plane2, plane3);

}






/*************************************************************************
*
*  function returns TRUE if an arrangement of 'a3_Planes' is FULL
*
*************************************************************************/
bool a3_Plane::isFull()
{
  if (planeList.size() > maxPlaneCount-1)
    return true;

  return false;
}


/*************************************************************************
*
*  function returns TRUE if an arrangement of 'a3_Planes' is EMPTY
*
**************************************************************************/
bool a3_Plane::isEmpty()
{
  if (planeList.size() == 0)
    return true;

  return false;
}



/*************************************************************************
*
*  function returns the number of entries in an arrangement of 'a3_Planes'
*
*************************************************************************/
int a3_Plane::getPlaneCount()
{
  return totalPlanes;
}



/****************************************************************************
 *
 *  function re-orders the 'listIndex' attribute for each plane in the
 *    arrangement
 *
 *  WHY DO WE NEED THIS FUNCTION?
 *  Recall that each object of type 'aPlane' contains an enumerated attribute
 *  called 'listIndex'. This number corresponds to the ordinal placement of
 *  the object within the list of all planes that are in the arrangement.
 *
 *  Whenever the user deletes a plane, it creates a missing number in the
 *  sequence of planes. So we must re-order the list, starting at the
 *  deleted item.
 *
 *   But LEDA designed its 'list_item' attributes to be READ-ONLY, so we
 *   cannot just re-index the existing list. We must make a new entry for
 *   each item in the list, updating the 'listIndex' attribute for each
 *   entry, as required.
 *
 ***************************************************************************/
void a3_Plane::sortPlanes()
{
  int newIndex = 0;
  aPlane newPlane;
  list_item it = planeList.first();
  
  for (int newIndex = 0; newIndex < totalPlanes; newIndex++)
    {
      aPlane newPlane(planeList.contents(it).getPlane(),
		      planeList.contents(it).getLine(),
		      newIndex+1);

      planeList.append(newPlane);
      it = planeList.succ(it);
      planeList.pop();      
    }
  hasChanged = true;
  
  return;
}
  


/*************************************************************
*
* function deletes all entries from a list of planes
* updates the count to ZERO
* does not delete the list structure
*
*************************************************************/
void a3_Plane::clear()
{
   planeList.clear();
   totalPlanes = 0;
   hasChanged = true;
   return;
}   



/**************************************************************
 *  function returns a 2-D line in x-y, corresponding to the
 *   cross-sectional view of a 3-D plane
 *
 *  pre: input arg is a valid d3_plane 
 **************************************************************/
leda_line a3_Plane::line_From_3D_Points(leda_d3_point pt1, leda_d3_point pt2)
{
   // strip the x-y coordinates from the 3-D points
   leda_point pointA(pt1.xcoord(), pt1.ycoord());
   leda_point pointB(pt2.xcoord(), pt2.ycoord());

   // define a line in 2-D 
   leda_line xyLine(pointA, pointB);
        
   return xyLine;
}


/***************************************************************
 *
 *  function checks whether a given plane intersects an x-y
 *   plane (i.e. a vertical face)
 *
 *  takes a plane and a line (by reference)
 *
 *  post: if result is true, line 'ln1' is assigned the value
 *        of the line of intersection between the two planes
 *        and the function returns  1;
 *        otherwise returns 0
 *
 ***************************************************************/
int a3_Plane::intersectsFace(leda_d3_plane & inputPlane,
			     leda_d3_point & q, leda_d3_point &r)
{
 // make a vertical plane at z=0 from 3 non-collinear points in 3D space
   leda_d3_point pt1(0.0, 0.0, 0.0);
   leda_d3_point pt2(1.0, 0.0, 0.0);
   leda_d3_point pt3(1.0, 1.0, 0.0);
   
   leda_d3_plane face(pt1, pt2, pt3);
   
   leda_d3_point t1, t2;                              // variables for line of intersection
 
   if (face.intersection(inputPlane, q, r) == 1)      // plane intersects vertical face
   {
//      leda_d3_line tempLine(t1, t2);                // construct the line of intersection
//      ln1 = tempLine;
      return 1;                        
   }
   return 0;
}


/***************************************************************
 *
 *
 *   function checks whether a given plane is in x-y orientation
 *   such that for any point 't' in plane 'p',
 *       t.zcoord() = w  (where w is some constant value)
 *
 *
 *      ________________________________
 *     |                                |
 *     |                                |
 *     |  a plane in x-y space          |
 *     |  (this is a view of the face)  | 
 *     |                                |
 *     |                                |
 *     | _______________________________|
 *
 *  returns 1 if plane is in x-y space
 *  returns 0 if plane has a slope in z
 *  returns -1 if plane neither intersects with an xy-plane nor is
 *    parallel to the xy-plane
 *
 ********************************************************************/
int a3_Plane::is_xyPlane(const leda_d3_plane &pl)
{

   // make 3 non-collinear points in 3D space
   leda_d3_point pt1(0.0, 0.0, 0.0);
   leda_d3_point pt2(1.0, 0.0, 0.0);
   leda_d3_point pt3(1.0, 1.0, 0.0);

   leda_d3_point t1, t2;               // variables for line of intersection
   
   leda_d3_plane xy(pt1, pt2, pt3);

   switch (pl.intersection(xy, t1, t2)) 
   {
      case 0:             // planes do not intersect, so check if parallel
	 if (pl.parallel(xy))
	    return 1;              // planes are parallel
	 else
	    return -1;              // should NEVER get this
	 break;

      case 1:
	 return 0;                  // planes intersect, thus
	 break;                     // the given plane is not in x-y orientation

      case 2:
	 return 1;                  // planes are identical
	 break;
   }
   return -1;                       // should never get this far
}
   

/***************************************************************
 *
 *
 *   function checks whether a given plane is horizontal
 *   such that for any point 't' in plane 'p',
 *       t.ycoord() = w  (where w is some constant value)
 *
 *  returns 1 if plane is horizontal
 *  returns 0 if plane has a slope in y
 *  returns -1 if plane neither intersects the horizontal plane
 *    nor is parallel to it
 *
 ********************************************************************/
int a3_Plane::is_horizPlane(const leda_d3_plane &pl)
{

   // make 3 non-collinear points in 3D space
   leda_d3_point pt1(0.0, 0.0, 0.0);
   leda_d3_point pt2(1.0, 0.0, 0.0);
   leda_d3_point pt3(1.0, 0.0, 1.0);

   leda_d3_point t1, t2;               // variables for line of intersection
   
   leda_d3_plane horiz(pt1, pt2, pt3);
   
   switch (pl.intersection(horiz, t1, t2)) 
   {
      case 0:             // planes do not intersect, so check if parallel
	 if (pl.parallel(horiz))
	    return 1;              // plane is parallel to horizontal
	 else
	    return -1;              // should NEVER get this
	 break;

      case 1:
	 return 0;                  // planes intersect, thus
	 break;                     // the given plane is not horizontal

      case 2:
	 return 1;                  // planes are identical
	 break;
   }
   return -1;                       // should never get this far
}

/***************************************************************************
 *
 *  function generates a random point in 3D space
 *
 *  Although LEDA documentation seems to indicate that a random 3D point
 *  generator is available, it is not in this library.
 *
 *  So we generate random integers and convert them to a 3D point
 *  based on 'doubles'.
 *
 *
 ****************************************************************************/
leda_d3_point a3_Plane::random_3d_point()
{
   int a,b,c;
   double x,y,z;
   
   leda_random_source randGen(-100, 100);       //  a random generator

   randGen >> a >> b >> c;
   randGen >> x >> y >> z;

   x = ((double)a) * x;
   x = (x <= 0)? x - xOffset: x + xOffset;
   
   y = ((double)b) * y;
   y = (y <= 0)? y - yOffset: y + yOffset;

   z = ((double)c) * z;
   z = (z <= 0)? z - (sqrt(pow(xOffset,2) + pow(yOffset,2))): z +  (sqrt(pow(xOffset,2) + pow(yOffset,2)));

   leda_d3_point r(x, y,z);
   
   return r;
}



/***********************************************************************
 * 
 *  write list of planes to text file with a file header 
 *  takes output stream by reference from calling function
 * 
 *
 **********************************************************************/
bool a3_Plane::savePlanesToFile(ofstream& outFile)
{
   list_item it;
   it = planeList.first();

   outFile << "#  Generated by ArrangePak-3D" << endl
	   << "#  Arrangement of Planes in 3D space" << endl
	   << "#  DO NOT EDIT THIS FILE!" << endl
	   << "#  Planes are defined by Point-Normal notation" << endl
	   << "#  --------------------------------------------" << endl
	   << "#     Point                            Normal Vector" << endl;
   
   int planeNum = 0;
   
   forall_items(it, planeList) {
     
     planeNum++;
     
     leda_d3_point tempPt(planeList.inf(it).getPlane().point1());
     leda_d3_point offsetPt(tempPt.xcoord()-xOffset, tempPt.ycoord()-yOffset, tempPt.zcoord());
     
     outFile << " " << planeNum << "  ";           // print the Plane Index Number
     outFile << offsetPt << "   \t"                // print the Point in the Plane
	     << ((planeList.inf(it).getPlane()).normal()).norm() << endl;     // print Unit Vector

   }   
   outFile.close();
   return true;
}




/************************************************************
 *
 *  reads list of planes from formatted text file
 *
 *  does some minor error checking
 *
 *  expects each line of the file to contain point-vector
 *   notation for a plane in 3D space
 *
 *  tests planes for validity, must intersect the viewport
 *   and be visible in our coordinate system
 *
 *  if operation succeeds, display will show the arrangement
 *  otherwise, window will be blank
 *
 ************************************************************/
bool a3_Plane::readPlanesFromFile(ifstream& inFile)
{
  planeList.clear(); 
  totalPlanes=0;
  bool result = true;

  int lineLength(60);
  char discard[lineLength];

  leda_d3_point q, r;
  leda_vector normVec;
  leda_d3_plane ppp;

  leda_list<leda_point> pointList;

  int c, planeNum;

  // discard the file header
  for (int headline = 0; headline < 6; headline++)
    inFile.getline(discard, lineLength);

  while (! inFile.eof())
    {      
      if ((c= inFile.peek()) == EOF) 
	break;
      
      // read in PLANE in POINT-NORMAL VECTOR notation from file

      inFile >> planeNum;
      inFile >> q;     
      if (inFile.fail()) {    // if FAILBIT set for another reason, then error
	result = false;
	break;
      }
      
      leda_d3_point offsetPt((q.xcoord()+xOffset), (q.ycoord()+yOffset), q.zcoord());
      //leda_d3_point offsetPt(q);

      inFile >> normVec;      // must read a VECTOR to match the previous POINT 
      if (inFile.fail()) {
	result = false;
	break;
      }

      while (inFile.get() != '\n')    // get rid of rest of line
	continue;

      if (!inFile.eof()){

	// construct a plane from the point and vector
	leda_d3_plane newPlane(offsetPt, normVec);  

	// make sure the plane is valid for this application
	if (intersectsFace(newPlane, q, r ) == 1 )  
	  {
	    leda_d3_line line1(q,r);
	    
	    leda_line line2;                // line2 will be the 2D line of intersection between plane and circle
	    line1.project_xy(line2);     
	    leda_circle circ(originPt, mainCircleRadius);
	    pointList = circ.intersection(line2);
	    
	    if (pointList.size() == 2) {
	      // add plane to list of planes, include corresponding line, and increment the count
	      
	      leda_line newLine(pointList.front(), pointList.back());
	      aPlane result(newPlane, newLine, ++totalPlanes);  // increment 'totalPlanes' first, vector is zero-indexed
	      planeList.append(result);
	      hasChanged = true;
	      
	    }
	    else {
	      result = false;
	      break;
	    }
	  }
	else {
	  result = false;
	  break;
	}
     }
    }    
  inFile.close();
  
  if (result == false) {     // if there were errors, reset everything
    planeList.clear();
    totalPlanes = 0;
  }
  
  return result;
}

    

/********************************************************************
 *
 *  function parses a list of 3D planes and generates output
 *   defined by parameter 'outputType' where 'outputType'
 *   corresponds to enumerated labels (see header file for this class)
 *   (e.g. generate graph, GLuskap file or node list file)
 *
 *  if graph is specified, it is an undirected graph
 *  if GLuskap file is specified, it is an undirected graph
 *  if node list file is specified, it includes each node's constituent planes
 * 
 *
 *  function returns function status by value  0 = success, non-zero if failure
 *
 *   There are SIX possible error cases:
 *    
 *      (1) any two planes in the arrangement are coincident (P1 = P2)
 *      (2) any two planes are exactly parallel
 *      (3) any three planes intersect to form a line
 *      (4) distance between any two vertices in the graph = 0  (i.e. there exist two identical vertices)
 *      (5) output graph missing some nodes
 *      (6) specified 'outputType' is unknown
 *
 *   In these cases, we do not construct a graph, but we do 
 *   pass back an error code
 *
 ********************************************************************/
int a3_Plane::extract_Graph(int outputType, const leda_string &outputFileName,
			    int &badPlane1, int &badPlane2, int &badPlane3)
{
   int i, j, k;
   list_item it, it2, it3;
   leda_d3_point linePt1, linePt2, iPt;
   float t;
   leda_string fileType;

   int degenCase = 0;                 // zero = no errors; non-zero = degenerate arrangement 
   const int badArgError = 6;         // incorrect argument for 'outputType'
   int functionStatus = 0;

   GRAPH<graphNode, int> G;           // initialize an empty (dynamic) parameterized graph
   
   leda_list<parameterNode> nodeList;        // node subsets, for walking lines of intersecting planes

   leda_list<parameterNode> outputNodeList;  // keep node data for output files

   bool bool_outputGraph = false;
   bool bool_outputGLuskap = false;
   bool bool_outputNodeList = false;

   switch (outputType) {         // set only one flag, corresponding to outputType
     case ledaGraph:
       bool_outputGraph = true;
       break;

     case gluskapFile:
       bool_outputGLuskap = true;
       break;

     case nodeFile:
       bool_outputNodeList = true;
       break;

     default:
       return badArgError;         // incorrect argument for 'outputType'
       break;
     }

   int nodeCounter = 0;


   /***********************************************
       begin by finding points in 3D space where 
       exactly three planes intersect
    
       each such point is a vertex (node) for the output graph
      
       load all the vertices into the graph, nodes are
        type 'graphNode'
    
       each vertex specifies exactly which three planes
        intersected to create that particular vertex
    
   ************************************************/

   it = planeList.first();
   for (i=0; i < planeList.size()-2; i++) {
     it2 = planeList.succ(it);
     for (j = i+1; j < planeList.size()-1; j++) {
       it3 = planeList.succ(it2);
       for (k = j+1; k < planeList.size(); k++) {

	  if ((((planeList.inf(it)).getPlane()).intersection(planeList.inf(it2).getPlane(), linePt1, linePt2)) == 1)
	  {
	   if ((((planeList.inf(it3)).getPlane()).intersection(linePt1, linePt2, iPt)) == 1)
	     {
	       //	       leda_string id("%i.%i.%i", i+1, j+1, k+1);  // zero-based planes are offset by +1 for human readability      
	       //      if (bool_outputNodeList)		 
	       //	 outFile << " " << nodeCounter << "  " << id << "  " << iPt << endl;
	       nodeCounter ++;
	       
	       graphNode gNode(i,j,k);
	       G.new_node(gNode);

	       // we shift the point coordinates by x- and y-Offset values for the viewport
	       leda_d3_point offsetPt(iPt.xcoord()-xOffset, iPt.ycoord()-yOffset, iPt.zcoord());

	       // save the node info to a list for use by output file
	       parameterNode outputFileNode(gNode, offsetPt, planeList.inf(it3).getPlane());
	       outputNodeList.append(outputFileNode);
       
	     }
	   //   else cerr << "inner loop result ==> " << ((planeList.inf(it3)).getPlane()).intersection(linePt1, linePt2, iPt) << endl;
	  }
	  //  else cerr << " outer loop result ==> "
	  //	  << ((planeList.inf(it)).getPlane()).intersection(planeList.inf(it2).getPlane(), linePt1, linePt2) << endl;
	  
	  //	  cerr << "incrementing it3\n";
	  it3 = planeList.succ(it3);
       }       
       //       cerr << "incrementing it2\n";
       it2 = planeList.succ(it2);
     }
     //     cerr << "incrementing it1\n";
     it = planeList.succ(it);
   }


   // verify that the number of nodes in the output graph is N-choose-3 where N is # of planes
   if (nodeCounter != (planeList.size() * (planeList.size()-1) * (planeList.size()-2) / 6))
   {
      cerr << "WARNING!  Node list has " << nodeCounter << " nodes in it.\n";
      cerr << "The node list is incomplete, probably due to parallel or coincident planes. " << endl;

      // insufficient number of nodes, so set degenCase = 5
      // calls to degeneratePlanes() and sort(nodeList) functions
      // should return the specific reason for the error, but
      // for now we at least know that the graph is rejected
      degenCase = 5;       
      
   }
   
   
   /**************************************************************
      Now we calculate all edges for the graph by 'walking' every
      line where two planes meet, and getting information about
      all other planes that intersect that line.
      The graph edges are the segments between adjacent planes 
      that intersect the line.
    
      Algorithm: for every pair of planes, if they intersect 
        at a line in space, then for all other planes, test
        whether they intersect the given line to form a point.
        For each plane that does intersect the line, generate
        a new vertex for the graph.
    
        Choose one such vertex, and sort all others by relative
        distance from the chosen vertex. Construct edges between 
        adjacent vertices.
    
        We consider four degenerate cases:
    
        (1) any two planes in the arrangement are coincident (P1 = P2)
        (2) any two planes are exactly parallel
        (3) any three planes intersect to form a line
        (4) distance between any two vertices in the graph = 0

        In these cases, we do not construct a graph, but we do 
        pass back an error code corresponding to the degenerate case.
        
   **************************************************************/

   it = planeList.first();
   for (i=0; i < planeList.size()-1; i++) {
     it2 = it;
     for (j = i+1; j < planeList.size(); j++) {
       nodeList.clear();
       it2 = planeList.succ(it2);
       it3 = planeList.first();
       for (k = 0; k < planeList.size(); k++) {

//	  cerr << " \n\n Comparing planes " << i+1 << ", " << j+1 << ", " << k+1 << endl;
	 if (k == i || k == j)   // the 'k'-th plane can't be the 'i'-th  or 'j'-th plane
	    ;
	 
	 
	 // check for degenerate case 1, 2 or 3
	 else if (degeneratePlanes(planeList.inf(it).getPlane(),
				   planeList.inf(it2).getPlane(),
				   planeList.inf(it3).getPlane(),
				   degenCase)) {
/*
            cerr << "Found bad planes while testing " << i+1 << ", " << j+1 << ", " << k+1 << " \n";
	    cerr << planeList.inf(it).getPlane() << endl;
	    cerr << planeList.inf(it2).getPlane() << endl;
	    cerr << planeList.inf(it3).getPlane() << endl;
*/
	       
	   badPlane1 = i+1;    // pass info back thru global variables
	   badPlane2 = j+1;    // for error reporting to user
	   badPlane3 = k+1;    // then quit the function 
	   return (degenCase);
	 }

	 else {
	    
	    
	    if ((((planeList.inf(it)).getPlane()).intersection((planeList.inf(it2).getPlane()), linePt1, linePt2)) == 1)
	       if ((((planeList.inf(it3)).getPlane()).intersection(linePt1, linePt2, iPt)) == 1)
	       {
		  leda_string id("%i.%i.%i", i,j,k);
		  //		 cout << " Found node during walk. ID # " << id << " at " << iPt << endl;
		  
		  graphNode tempNode(i,j,k);
		  parameterNode pNode(tempNode, iPt, planeList.inf(it3).getPlane());
		  nodeList.append(pNode);      // append node info to temporary list
		  
	       }
	 }
	 it3 = planeList.succ(it3);
	 
       }
/*
  cerr << "\n\nHave tested for degenerate arrangement on all planes, and found none." << endl
	    << "This number MUST be a ZERO ==>  " << degenCase << endl << endl;
*/     
       if (degenCase != 0) 
	  return degenCase;
       
       // sort the array of nodes from one end to the other
       // sorted array will be ordered by adjacency in 3D space
       
       degenCase = sort(nodeList);
       
       if (degenCase != 0)    // test whether two nodes have same coordinates    
	  return degenCase;
       
       
       extractEdgesToGraph(nodeList, G);   // copy edges on this line to the Graph
     }
     
     it = planeList.succ(it);
   }
   
   
   // got this far, so no degenerate case found for the graph
   // thus, we construct the output graph or GLuskap file
   
   if (bool_outputGLuskap) {
     leda_string fileName = outputFileName;
     fileType = ".mg";
     fileName += fileType;     
     writeGLuskapFile(outputNodeList, G, fileName);
   }

   
   if (bool_outputNodeList) {
     leda_string fileName = outputFileName;
     fileType = ".nodes";
     fileName += fileType;
     writeNodeFile(outputNodeList, fileName);
   }

   if (bool_outputGraph) {
     GraphWin gw(G, "ArrangePak-3D");
     
     gw.set_node_color("red", true);        // setup the visual graph
     gw.set_edge_color("navy", true);
     gw.set_edge_direction(undirected_edge, true);
     
     gw_random_embed(gw);                  // start with random layout of nodes
     gw.display(leda_window::center, leda_window::center);  
     gw.edit();
   }

   return 0;    // no degenerate cases in the arrangement

}



/********************************************************
 *  compare two planes in 3D space
 *
 *  input parameters: two structures of type parameterNode
 *                    by reference
 *     
 *  given a point in space, returns a value 'n' corresponding
 *   to which side of a given plane the point is on
 *
 *  calls  LEDA function 'side_of' to calculate the result
 *
 *
 ********************************************************/
int comparePlanes(const parameterNode &p1, const parameterNode &p2)
{
  int n = (p1.pl).side_of(p2.pt);
  return n;
}



/********************************************************
 *  test three planes in 3D space
 *
 *  if any of the following cases are true, the arrangement
 *   is a degenerate case, and we return an error code
 *   by reference, as well as boolean TRUE
 *     
 *  (1) any two planes in the arrangement are coincident
 *  (2) any two planes are exactly parallel
 *  (3) any three planes intersect to form a line
 * 
 *  if three planes meet in a point, then no problem
 *   found so we return boolean FALSE, and the error 
 *   code is zero (non-error)
 *
 ********************************************************/
bool a3_Plane::degeneratePlanes(const leda_d3_plane &P1, 
				const leda_d3_plane &P2,
				const leda_d3_plane &P3, int &errCase) {
  int status;
  leda_d3_point q, r, s;
  
  errCase = 0;
  // test for coincident planes
  if (P1 == P2 || P1 == P3 || P2 == P3) {
    errCase = 1;  
    return true;
  }

  // test for parallel planes
  if (P1.parallel(P2) || P1.parallel(P3) || P2.parallel(P3)) {
    errCase = 2;
    return true;
  }

  // test for case 3
  status = P1.intersection(P2, q, r); 

  if (status == 0) {         // two planes don't meet so
    if (P1.parallel(P2)) {   // double-check, this MUST be true
      errCase = 2;
      return true;
    }
    else {                 // if we get here, there is a BIG problem
      cerr << "non-intersecting planes not parallel... bailing out\n";
      exit (-1);
    }
  }
  
  else if (status == 1)       // most times, this will be true
    {                         // got here, so 2 planes meet on a line

      status = (P3.intersection(q, r, s)); 

      if (status == 2) {      // if third plane contains the line
	 errCase = 3;          // error corresponds to degenerate case 3
	 return true;          // such that all 3 planes intersect on a line
      }

      else if (status == 1)    // three planes meet in a point, so there
	 return false;         // is no degenerate case; return success 
    }
  
  
  
  status = P1.intersection(P3, q, r);
  if (status == 0) {         // two planes don't meet
    if (P1.parallel(P3)) {   // double-check, this MUST be true
      errCase = 2;
      return true;
    }
    else {                 // if we get here, there is a BIG problem
      cerr << "non-intersecting planes not parallel... bailing out\n";
      exit (-1);
    }
  }
  
  // don't check if all three meet on a line, did that already

  status = P2.intersection(P3, q, r);
  if (status == 0) {         // two planes don't meet
    if (P2.parallel(P3)) {   // double-check, this MUST be true
      errCase = 2;
      return true;
    }
    else {                 // if we get here, there is a BIG problem
      cerr << "non-intersecting planes not parallel... bailing out\n";
      exit (-1);
    }
  }

  return false;          // got this far, no degenerate case found
  
}


/**************************************************************
 * sort a parameterized list (of nodes) by distance
 *  from an arbitrary reference node (the first element in list)
 *
 * in case two nodes in the list have relative distance = 0
 *  then the arrangement is a degenerate case, and we return 
 *  a non-zero result by value, corresponding to failure
 *
 *  otherwise return the node list, sorted by distance
 *   thus adjacent nodes in the list share an edge in the graph
 *
 **************************************************************/
int a3_Plane::sort(leda_list<parameterNode> &nodeList)
{

  if (nodeList.empty())        // if list is empty, quit early
    return 0;

  leda_list<parameterNode> leftList;
  leda_list<parameterNode> rightList;
  parameterNode refNode = nodeList.pop();

  int success = 0;   // return this if degenerate case not found
  
  int compResult;
  
  list_item itA = nodeList.first();  

  while (itA != nil)
    {
      compResult = comparePlanes(refNode, nodeList.inf(itA));

      switch (compResult)
	{
	case -1:                           // comparison plane is "left" of reference
	  leftList.append(nodeList.inf(itA));
	  break;
	  
	case 1:                           // comparison plane is "right" of reference
	  rightList.append(nodeList.inf(itA));
	  break;
	  
	case 0:        // found two nodes with distance = 0, so quit early
	  return 4;    // error code corresponds to 'degenerate case 4' 
	  break;
	  
	default:       // should NEVER get here
	  cerr << "confused by errors while sorting node list, bailing out\n";
	  exit (-1);
	  
	}
      
      itA = nodeList.succ(itA);
    }


  if (leftList.empty())
    ; // There are no planes to the LEFT of the REFERENCE plane
  else
    sortByDistance(refNode, leftList);
  
  if (rightList.empty())
    ; // There are no planes to the RIGHT of the REFERENCE plane
  else
    sortByDistance(refNode, rightList);
  
  nodeList.clear();

  
  // rebuild the overall list in order from end-to-end, so we
  // must reverse the first list 
  // (thank-you E.N. & S.W. for identifying and resolving this bug !!)

  leftList.reverse();
  while (! leftList.empty())
    nodeList.append(leftList.pop());
  
  nodeList.append(refNode);
  
  while (! rightList.empty())
    nodeList.append(rightList.pop());
  
  return success;
}




/**************************************************************
* 
*  function sorts a list of type 'parameterNode' in order by distance 
*   from a given point on a line
*
*  takes a 'referenceNode' and a list called 'splitList' by reference
*
*  compares each 'parameterNode' to the 'referenceNode' and
*   sorts the list using the 'pt' attribute for the node
*
*  (see the description of a 'parameterNode' datatype defined
*   elsewhere in this source code)
*
*
**************************************************************/
void a3_Plane::sortByDistance(const parameterNode & refNode, leda_list<parameterNode> &splitList)
{
  list_item it1, it2;
  
  parameterNode tempNode;
  
  for (it2 = splitList.succ(splitList.first()); 
       it2 != nil; 
       it2 = splitList.succ(it2))
    {
      tempNode = splitList.inf(it2);
      
      for (it1 = it2; 
	   splitList.pred(it1) != nil &&
	     (refNode.pt).distance(tempNode.pt) < (refNode.pt).distance(splitList.inf(splitList.pred(it1)).pt);
	   it1 = splitList.pred(it1))
	
	splitList.assign(it1, splitList.inf(splitList.pred(it1)));
      
      splitList.assign(it1, tempNode);
    }

  return;
}



/**************************************************************
* 
*  function generates an undirected edge for each pair of 
*   nodes belonging to a line of intersection for two planes
*
**************************************************************/
void a3_Plane::extractEdgesToGraph(leda_list<parameterNode> &nodeList, GRAPH<graphNode, int> &G)
{
  // need to identify the planes adjacent to any given point
  // so that if two points share two planes, they are connected by an edge
  
  list_item listIt1, listIt2;
  NodeIt nIt1(G), nIt2(G);
  

  if (nodeList.size() > 1)
    {
      // for the entries in the main node array  (ie. nodeList)
      // find the 'k'-th node in the BigList (G<node>) and return it via nIt1;
      // find the 'k+1'-th node in the BigList (G<node>) and return it via nIt2;

       listIt1 = nodeList.first();

       for (int k=0; k < nodeList.size()-1; k++) {	
	  listIt2 = nodeList.succ(listIt1);
	
	  nIt1.reset();
	  nIt2.reset();
	
	  while (nIt1.valid() && G.inf(nIt1.get_node()) != nodeList.inf(listIt1).gN)
	  {
	     ++nIt1;
	  }

	  
	while (nIt2.valid() && G.inf(nIt2.get_node()) != nodeList.inf(listIt2).gN)
	{
	   ++nIt2;
	}

	G.new_edge(nIt1.get_node(), nIt2.get_node());

	listIt1 = nodeList.succ(listIt1);
      }
    }
  return;
}





/*********************************************************************
 *
 * function exports node and edge lists to a GLuskap file format
 * output data includes the number of nodes and edges in the graph,
 *  followed by the node list with 3D coordinates with RGB colors,
 *  followed by edge list giving source and target nodes, with RGB colors
 * 
 *
 ********************************************************************/
void a3_Plane::writeGLuskapFile(leda_list<parameterNode> &gluskapNodeList,
				GRAPH<graphNode, int> &G,
				leda_string &fileName)
{
  int nodeCount = 0;
  list_item nodeIter = gluskapNodeList.first(); 

  leda_string edgeRGBValues("   1  1  1");       // rgb values for GLuskap edge color 1,1,1=>white

  ofstream outFile;
  outFile.open(fileName);
    
  outFile << G.number_of_nodes() << "   " << G.number_of_edges() 
	  << "   none  0  0  " << endl;


  // output the list of nodes to GLuskap format file
  while (nodeIter != nil) {
    outFile << nodeCount << "    " << gluskapNodeList.inf(nodeIter).pt.xcoord() 
	    << "    " << gluskapNodeList.inf(nodeIter).pt.ycoord()
	    << "    " << gluskapNodeList.inf(nodeIter).pt.zcoord()
	    << "     none  1  0  0 " << endl;
    
    nodeCount ++;
    nodeIter = gluskapNodeList.succ(nodeIter);
  }


  NodeIt nodeIt(G);
  EdgeIt edgeIt(G);
  nodeCount = 0;
 
  // output the edge information for all edges
  while (edgeIt.valid()){

    // output the source node label
    while (nodeIt.valid() && (nodeIt.get_node() != source(edgeIt.get_edge())))
      {
	++nodeIt;
	++nodeCount;
      }
    outFile << nodeCount;


    // output the target node label
    nodeIt.reset();
    nodeCount = 0;
    while (nodeIt.valid() && (nodeIt.get_node() != target(edgeIt.get_edge())))
      {
	++nodeIt;
	++nodeCount;
      }
    outFile << "  " << nodeCount;
    
    outFile << "  " << edgeRGBValues << endl;

    ++edgeIt;
    nodeIt.reset();
    nodeCount = 0;
    
  }  // done looping through the list of edges
  
  return;

}








/*********************************************************************
 *
 * function writes the list of vertices to a file
 * output data includes the zero-based index of the vertex,
 *  followed by the constituent parent planes that intersect
 *  to form the vertex in 3D space, followed by the coordinates
 *  of the node in 3D space
 *
 ********************************************************************/
void a3_Plane::writeNodeFile(leda_list<parameterNode> &outputNodeList,
			     leda_string &fileName)
{
  int nodeCount = 0;
  list_item nodeIter = outputNodeList.first(); 
  
  ofstream outFile;
  outFile.open(fileName);
  outFile << "# NODE LIST FOR GRAPH - GENERATED by ArrangePak-3D " << endl;
  
  while (nodeIter != nil) {
    outFile << " " << nodeCount << "  "
	    << outputNodeList.inf(nodeIter).gN.a+1 << "."
	    << outputNodeList.inf(nodeIter).gN.b+1 << "."
	    << outputNodeList.inf(nodeIter).gN.c+1 
      	    << "   ( " << outputNodeList.inf(nodeIter).pt.xcoord() 
	    << ",  " << outputNodeList.inf(nodeIter).pt.ycoord()
	    << ",  " << outputNodeList.inf(nodeIter).pt.zcoord()
	    << " )" << endl;
    
    nodeCount ++;
    nodeIter = outputNodeList.succ(nodeIter);
  }

  return;

}





/***********************************************************************
 * if current arrangement has been saved to a file, this function
 *  returns BOOLEAN TRUE; otherwise returns BOOLEAN FALSE
 *
 *
 ***********************************************************************/
bool a3_Plane::isSaved()
{
  if (!hasChanged)
    return true;
  return false;
}



istream& operator >>(istream &is, const aPlane &a)
{    
  is >> "\n";      //  should do more than this but don't need to 
  return is;
}



ostream& operator <<(ostream &os, const aPlane &a)
{
  os << a.getPlane() << endl;
  return os;
}


void operator <<(ifstream& iFile, const aPlane &a)
{
   return;
}



void operator <<(ofstream& oFile, const aPlane &a)
{
   oFile << a.getPlane() << endl;
   return;
}


ofstream& operator<<(ofstream& oFile, leda_d3_plane &d)
{
   oFile << d << endl;
   return oFile;
}


