/****************************************************************
 *
 *
 *   a3_Plane.h
 *
 *   An arrangement of planes, as a class.
 *   Used for ArrangePak-3D.
 *
 *
 *   ArrangePak-3D 
 *   version 1.04
 *   original version:   October, 2003
 *   this revision:      May 2004
 *
 *
 *   This is the class library for an arrangement of planes in 3-dimensional
 *   space. An arrangement of planes maintains a list of planes in 3-space,
 *   a 'viewLine' attribute that represents the intersection of the given
 *   plane with a viewport at 'z'=0 in the cartesian coordinate system,   
 *   and additional member attributes such as 'totalPlanes', the number of
 *   planes in the arrangement, as well as necessary member functions.
 *
 *------------------------------------------------------------------------
 *
 *   copyright (c) 2004, 2003 Stephen Wismath, Ph.D.
 *   University of Lethbridge
 *   Lethbridge, AB   CANADA
 *
 *------------------------------------------------------------------------
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *------------------------------------------------------------------------
 *
 *   Licence terms do not apply to the Library of Efficient Data Types and 
 *   Algorithms, which is licensed separately by 
 *       Algorithmic Solutions Software Gmbh 
 *       http://www.algorithmic-solutions.com/
 *
 *************************************************************************/
#ifndef _PLANE_ARRANGEMENT_
#define _PLANE_ARRANGEMENT_

#include <LEDA/line.h>               // includes LEDA::point
#include <LEDA/d3_plane.h>           // includes LEDA::d3_line and LEDA::d3_point
#include <LEDA/graph.h>
#include <LEDA/list.h>               // required for LEDA::list class
#include <LEDA/window.h>             // required for LEDA::window operations
#include <LEDA/memory.h>             // for LEDA memory manager
//#include <LEDA_edge_array.h>         // for exporting edges to GLuskap

#include <fstream>

#include "graphNode.h"
#include "parameterNode.h"
#include "aPlane.h"


enum {ledaGraph, gluskapFile, nodeFile};   // labels define what kind of output to generate

const int minPlaneCount = 1;
const int maxPlaneCount = 20;         // arbitrary maximum number of planes in the arrangement
const int defaultQuan = 3;            // arbitrary default number of planes to start with


class a3_Plane 
{

 private:
  
  leda_list<aPlane> planeList;       // leda_list is implemented as a list of pointers to containers
  bool hasChanged;                   // false if current arrangement is not saved to a file
  int totalPlanes;
  
  
  bool isFull();                     // test whether arrangement has maximum allowable planes
  bool isEmpty();                    // test whether arrangement is the empty set
  

  // line along x-y corresponds to cross-section view of plane
  leda_line line_From_3D_Points(leda_d3_point, leda_d3_point);


  // tests whether plane intersects the plane at (x, y, 0);
  int intersectsFace(leda_d3_plane &, leda_d3_point&, leda_d3_point &);

  // tests whether orientation of the plane is in (x,y, 0)
  int is_xyPlane(const leda_d3_plane &pl);
  
  // tests whether plane is in (x, 0, z)
  int is_horizPlane(const leda_d3_plane &pl);
  
  // generate a random point in 3D space
  leda_d3_point random_3d_point();
  
  int sort(leda_list<parameterNode> &nodeList);

  void extractEdgesToGraph(leda_list<parameterNode> &nList, GRAPH<graphNode, int> &G);

  // extract graph information (nodes, edges) from arrangement
  int extract_Graph(int, const leda_string &, int &, int &, int &);

 
  void writeGLuskapFile(leda_list<parameterNode> &, 
			GRAPH<graphNode, int> &, leda_string &);

  void writeNodeFile(leda_list<parameterNode> &outputNodeList,
		     leda_string &fileName);

  bool degeneratePlanes(const leda_d3_plane &P1, 
			const leda_d3_plane &P2,
			const leda_d3_plane &P3, int &errCase);

  
  
 public:
  
  a3_Plane();
  ~a3_Plane();
  
  // add random plane to arrangement
  int add_Random_Plane();

  // add default plane to arrangement
  int add_Default_Plane();

  // delete a plane
  int delete_Plane(double x, double y);


  void swapPlane(leda_d3_plane&, int);

  leda_line get_2D_Line(int i);
  leda_d3_segment get_XY_Axis(int);
  leda_d3_segment get_3D_Axis(int, const leda_point&, const leda_point&, int);


  int getPlane(int, leda_d3_plane &);
  int findPlane(double, double);
  
  void sortPlanes();
  leda_d3_point getRotationPoint(int, int);
  
  int to_Graph(const int, const leda_string &, int &, int &, int &);
  int getPlaneCount();
  void clear();

  bool savePlanesToFile(ofstream&);
  bool readPlanesFromFile(ifstream&);
  /*
  int sort(leda_list<parameterNode> &nodeList);
  void extractEdgesToGraph(leda_list<parameterNode> &nList, GRAPH<graphNode, int> &G);
  */
  bool isSaved();

  void sortByDistance(const parameterNode & refNode, leda_list<parameterNode> &splitList);

  friend istream& operator >>(istream &is, const aPlane &a);
  friend ostream& operator <<(ostream &os, const aPlane &a);
  
  friend void operator <<(ifstream& iFile, const aPlane &a);
  friend void operator <<(ofstream& oFile, const aPlane &a);
  friend ofstream& operator <<(ofstream& oFile, leda_d3_plane &d);
  
  LEDA_MEMORY(a3_Plane)
    
};
  
int comparePlanes(const parameterNode&, const parameterNode&);

#endif  

