/*************************************************************************
 * 
 *   a3d.cpp
 *
 *   driver for ArrangePak-3D, a tool for analyzing arrangements of planes
 *   in three dimensions
 *
 *   ArrangePak-3D 
 *   version 1.04
 *   original version:   October, 2003
 *   this revision:      May 2004
 *
 *   This application supports analysis of 3-dimensional planes by constructing
 *   a virtual viewing cylinder.
 *
 *   The cylinder provides a cross-sectional view of planes that intersect
 *   it, within the bounds of the cylinder. Calculations are made on the 
 *   entire plane (beyond the bounds of the viewing cylinder).
 *
 *------------------------------------------------------------------------
 *
 *   copyright (c) 2004, 2003 Stephen Wismath, Ph.D.
 *   University of Lethbridge
 *   Lethbridge, AB   CANADA
 *
 *------------------------------------------------------------------------
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *------------------------------------------------------------------------
 *
 *   Licence terms do not apply to the Library of Efficient Data Types and 
 *   Algorithms, which is licensed separately by 
 *       Algorithmic Solutions Software Gmbh 
 *       http://www.algorithmic-solutions.com/
 *
 *****************************************************************/

#include <X11/cursorfont.h>
#include <math.h>

#include <LEDA/graph.h>           // required for graph objects
#include <LEDA/menu.h>            // for window menu buttons
#include <LEDA/window.h>          // required for LEDA::window interface
#include <LEDA/graphwin.h>        // for GraphWin objects
#include <LEDA/d3_window.h>       // supports 3-D visualization 
#include <LEDA/vector.h>          // for 3D geometry vectors, NOT dynamic arrays    
#include <LEDA/circle.h>          // for the viewport circle
#include <LEDA/d3_line.h>           // for 3D lines
#include <LEDA/segment.h>         // for the viewport "moveable" lines
#include <LEDA/file.h>            // for file I/O
#include <LEDA/memory.h>          // for LEDA memory manager

#include "a3d_messages.h"         // user prompt and error message strings 
#include "a3d.h"                  // interface for this file 
#include "a3_Plane.h"             // plane arrangement class
#include "a3d_colors.h"           // declaration of colors for various window items

#define PI 3.141592654

char * APAK_PATH = new char[255];

leda_string filePath(get_directory());    // working directory

int screen_width, screen_height, window_width, window_height;

int defaultCursor = XC_left_ptr;  // initialize to standard cursor


// setup some variables for drawing the viewport, lines, etc

const double circleScaleFactor = 0.47;       // "viewport" radius = window width x 0.47

double mainCircleRadius;
const int xScale = 100;            // x scale extends in either direction by this amount
const int yScale = 100;            // y scale extends in either direction by this amount

double zVal = 150.0;               // initial value for the frontFace plane, but it will change

// construct a vertical plane somewhere between the display and the user
leda_d3_point frontPt1(0.0, 0.0, zVal);
leda_d3_point frontPt2(1.0, 0.0, zVal);
leda_d3_point frontPt3(1.0, 1.0, zVal);
leda_d3_plane frontFace(frontPt1, frontPt2, frontPt3);

leda_point originPt;               // this will be the coordinates of our universe origin
                                   // relative to the LEDA origin in the lower left corner

double xOffset, yOffset;           // these are used to shift the window origin from
                                   // lower left corner to center of window

a3_Plane pA;                       // an arrangement of planes in 3D space


const int maxSlider = 20;          // slider control for user to choose number of planes
const int sliderMaxPlanes = (maxPlaneCount > maxSlider) ? maxSlider : maxPlaneCount;

int planeQuantity = defaultQuan;   // default number of planes is defined elsewhere
int int_crosshair = 0;             // initial crosshair
bool bool_showCoord = true;        // default is SHOW the mouse coordinates
leda_string winMsg("");            // clear the status message in the display, top left corner



//------------------------------------------------------------------------
//************* FINALLY... THE MAIN PROGRAM STARTS HERE!! ****************
//------------------------------------------------------------------------
int  main() {

   leda_window startupWin(1,1);

   if (setupScreen(startupWin) != 0) {
      cerr << screenFailMsg << endl;
      return -1;
   }
  
  // initialize a square window for the graph
  leda_window planeWin(window_width, window_height, "ArrangePak-3D");
  planeWin.set_precision(32);

  leda_graph G;

  leda_node_array<leda_vector> nArray(G);


  //  displaySplashPanel();              // used in BETA version only
 
  // build user menu for the main window
  // **** most LEDA data types are prefixed by 'leda_'
  // **** but a LEDA::menu data type is not prefixed!! 
  int menuIndex = 0;
  menu fileMenu;
  menu planeMenu;
  menu optionMenu;
  menu crossHairMenu;
  menu cursorMenu;
  menu helpMenu;

  // initialize a slider and menu buttons
  planeWin.int_item("Planes:", planeQuantity, 1, sliderMaxPlanes);
  
  fileMenu.button("&Load", menuIndex++);
  fileMenu.button("&Save", menuIndex++);
  fileMenu.button("&Generate graph", menuIndex++);
  fileMenu.button("Export to GLus&kap file", menuIndex++);
  fileMenu.button("Generate &Node List", menuIndex++);
  fileMenu.button("&Exit", menuIndex++);
  planeWin.button("&File", fileMenu);

  planeMenu.button("Ge&nerate random planes", menuIndex++);
  planeMenu.button("Add &random plane", menuIndex++);
  planeMenu.button("Add &default plane", menuIndex++);
  planeMenu.button("De&lete planes", menuIndex++); 
  planeMenu.button("Dra&g planes", menuIndex++);
  planeMenu.button("&Show all planes", menuIndex++);
  planeWin.button("&Planes", planeMenu);

  optionMenu.button("C&rosshair", crossHairMenu);
  crossHairMenu.button("&Large", menuIndex++);
  crossHairMenu.button("&Small", menuIndex++);
  crossHairMenu.button("&None", menuIndex++);

  optionMenu.button("C&ursor", cursorMenu);
  cursorMenu.button("&Standard", menuIndex++);
  cursorMenu.button("45-deg Left &Arrow", menuIndex++);
  cursorMenu.button("&Light Cross", menuIndex++);
  cursorMenu.button("&Dark Cross", menuIndex++);
  cursorMenu.button("&X", menuIndex++);
  cursorMenu.button("Dot&box", menuIndex++);

  optionMenu.button("&Toggle Cursor Coordinates", menuIndex++);
  optionMenu.button("Set file &path", menuIndex++);
  planeWin.button("&Options", optionMenu);

  helpMenu.button("&Online help", menuIndex++);
  helpMenu.button("&About ArrangePak-3D", menuIndex++);   
  planeWin.button("&Help", helpMenu);

  planeWin.button("&Quit", menuIndex++);
  planeWin.make_menu_bar();

  // display the main window at horizontal center and vertical
  //  top offset by a small amount (arbitrarily set to 10)
  planeWin.display(leda_window::center, leda_window::max+10);

  refresh(planeWin, 0);  
  generatePlanes(pA, planeWin);
  refresh(planeWin, 0);

  
  /******************************************************
    Now iterate through a loop until user quits.
    The loop handles mouse events, button clicks and 
    window dragging / resizing.

    There are really two windows -- the graph window and the menu panel
    thus we must acquire information about the window as well as the event.
  *******************************************************/

  bool exitProgram = false;   // when 'exitProgram' == TRUE, quit the program  
  do {
    
/***************************************************************
 **  Enumerated in LEDA/incl/LEDA/impl/x_windows.h
 **
 **   Window Events              LEDA ENUM Constant
 **
 **       key_press_event              0
 **       key_release_event            1
 **       button_press_event           2
 **       button_release_event         3
 **       panel_press_event            4
 **       configure_event              5
 **       exposure_event               6
 **       motion_event                 7
 **       destroy_event                8
 **       timer_event                  9
 **       no_event                    10
 **       
 **
 ***************************************************************/
   
    leda_window *wPtr = NULL;        // pointer to window that mouse event occurs within
    int winNumber = 0;               // enumerated value corresponding to the window of the event
    int buttonValue;                 // enumerated value corresponding to menu or mouse button when clicked
    int winEvent = 0;                // enumerated value corresponding to the window event (see above)
    int eventValue = 0;
    double winX, winY;               // X- and Y- coordinates of the mouse when event occurs
    float t;

    // points that define line of intersection between two planes
    leda_d3_point i_pt1;
    leda_d3_point i_pt2;

    // plane in 3D space
    leda_d3_plane plane_3d;

    winEvent = read_event(wPtr, eventValue, winX, winY);
    
    if (wPtr == &planeWin) winNumber = 1;
    
    switch (winNumber)
      {
      case 1:
	switch (winEvent)
	  {
	  case key_press_event:        // do nothing for keyboard action
	  case key_release_event:
	    break;
	    
	  case configure_event:        // if user moves the window, refresh it
	  case panel_press_event:
	  case exposure_event:
	    refresh(planeWin, 0);
	    break;
	    
	  case motion_event:           // if flag is set, update cursor coordinates
	    if (bool_showCoord)
	      refreshCursorCoord(planeWin, winX, winY);    
	    break;
	    
	  case button_press_event:
	    buttonValue = eventValue;
	    
	    switch (buttonValue)
	      {
	      case 0:                  // load a 3D arrangement from a file
		/* 
		   if (pA.hasChanged)
		     planeWin.message(saveBeforeDelete);
		   else
		*/
		readFromFile(planeWin);
		redrawWindow(planeWin, 0);
		break;
		
	      case 1:           // save the arrangement to a file
		saveToFile(planeWin);
		break;
		
		
	      case 2:           // generate a LEDA graph from this arrangement
		generateOutput(planeWin, ledaGraph);
		break;
		
	      case 3:           // output file in GLuskap .mg file format
		generateOutput(planeWin, gluskapFile);
		break;

	      case 4:           // output file of nodes and their parent planes
		generateOutput(planeWin, nodeFile);
		break;

	      case 5:           // user terminates the program
		 exitProgram = true;  
		 break;
		
	      case 6: 		// generate a random plane arrangement
		generatePlanes(pA, planeWin);
		refresh(planeWin,0);
		break;
		
	      case 7:           // add a plane with random orientation
		refresh(planeWin,0);
		if (pA.getPlaneCount() == maxPlaneCount)
		  planeWin.acknowledge(noAdd);
		else {
		  addRandomPlane(pA, planeWin);
		  refresh(planeWin,0);
		}
		break;
		
	      case 8:           // add a plane in default orientation
		refresh(planeWin,0);
		if (pA.getPlaneCount() == maxPlaneCount)
		  planeWin.acknowledge(noAdd);
		else {
		  addDefaultPlane(pA, planeWin);
		  refresh(planeWin,0);
		}
		break;
		
	      case 9:           // delete lines from the arrangement 
		if (pA.getPlaneCount() == minPlaneCount)
		  planeWin.acknowledge(noDelete);
		else {
		  deletePlane(pA, planeWin);
		  pA.sortPlanes();
		}
		break;
		
		
	      case 10:          // drag planes
		if (pA.getPlaneCount() == 0)
		  planeWin.acknowledge(noPlanes);
		else
		  dragPlanes(pA, planeWin);
		break;
		

	      case 11:          // show ALL planes in "pseudo-3D"
		showAllPlanes(planeWin);
		redrawWindow(planeWin, 0);
		break;
		

	      case 12:
		int_crosshair = 0;
		redrawWindow(planeWin, 0);
		break;
		

	      case 13:
		int_crosshair = 1;
		redrawWindow(planeWin, 0);
		break;

		
	      case 14:
		int_crosshair = 2;
		redrawWindow(planeWin, 0);
		break;

		
	      case 15:
		defaultCursor = XC_left_ptr;
		planeWin.set_cursor(defaultCursor);
		refresh(planeWin, 0);
		break;

		
	      case 16:
		defaultCursor = XC_top_left_arrow;
		planeWin.set_cursor(defaultCursor);
		refresh(planeWin,0);
		break;		  

		
	      case 17:
		defaultCursor = XC_crosshair;
		planeWin.set_cursor(defaultCursor);
		refresh(planeWin, 0);
		break;

		
	      case 18:
		defaultCursor = XC_plus;
		planeWin.set_cursor(defaultCursor);
		refresh(planeWin, 0);
		break;

		
	      case 19:
		defaultCursor = XC_X_cursor;
		planeWin.set_cursor(defaultCursor);
		refresh(planeWin, 0);
		break;

		
	      case 20:
		defaultCursor = XC_dotbox;
		planeWin.set_cursor(defaultCursor);
		refresh(planeWin, 0);
		break;


	      case 21:
		bool_showCoord = (!bool_showCoord);
		redrawWindow(planeWin, 0);
		break;

		
	      case 22:
		setPath(planeWin);
		refresh(planeWin, 0);
		break;
		

	      case 23:
		 HelpFile("Apak3D_Help");
		 break;
		

	      case 24:
		displayAbout();
		break;
		

	      case 25:                    // choice = QUIT
		exitProgram = true;
		break;
		

	      case MOUSE_BUTTON(1):       // left mouse clicked in viewport
		planeQuantity = pA.getPlaneCount();    // refresh the SLIDER bar
                planeWin.redraw_panel();
		break;
		

	      case MOUSE_BUTTON(3):       // RIGHT MOUSE click opens floating menu
		planeWin.get_mouse(winX, winY);
		popUpMenu(pA, planeWin, winX, winY);
		refresh(planeWin,0);
		break;
		
	      }
	  }
      }
    
  } while (!exitProgram);
  
  
  // user exit from the loop, so close application
  // LEDA memory manager handles cleanup, nothing else to do (I think)
  return 0;
  
  
}



//***************  END OF MAIN PROGRAM  ***********************************
//
//***********  FUNCTION DEFINITIONS FOLLOW ********************************



/**************************************************************************
*
*  find out the screen resolution and set window sizes accordingly
*
*  the screen and window sizes are used by the arrangement graph 
*   window algorithms
*
*  pre: Takes LEDA::window as input by reference
*  post: returns value 0 if function succeeds
*
*  NOTE:
*  LEDA-4.1(Linux) allows us to call this function with no parameters,
*   make the window, get the screen size, and destroy the window when leaving
*   the function. This is a good approach, since we no longer require the window.
*
*  BUT...  LEDA-4.3 (UNIX) prevents us from doing the same. 
*   We must create a LEDA::window in the 'main' function BEFORE calling 
*   THIS function or else the application terminates with a BAD FONT error.
*
*  Thus, since we are forced to create the LEDA::window in 'main' before
*   calling THIS function, we use that window as a parameter for this
*   function, thus maintaining some semblance of clarity and coding elegance.
*
****************************************************************************/
int setupScreen(leda_window & screenSize)  
{  

   screenSize.display(leda_window::center, leda_window::center);

   if      (screenSize.xpos() < 321)  { screen_width =  640; screen_height =  480; }
   else if (screenSize.xpos() < 401)  { screen_width =  800; screen_height =  600; }
   else if (screenSize.xpos() < 513)  { screen_width = 1024; screen_height =  768; }
   else if (screenSize.xpos() < 577)  { screen_width = 1152; screen_height =  864; }
   else if (screenSize.xpos() < 641)  { screen_width = 1280; screen_height = 1024; }
   else if (screenSize.xpos() < 801)  { screen_width = 1600; screen_height = 1200; }
   else if (screenSize.xpos() < 901)  { screen_width = 1800; screen_height = 1440; }
   else if (screenSize.xpos() < 961)  { screen_width = 1920; screen_height = 1440; }
   else  { screen_width = 2048; screen_height = 1536; }
   
   screenSize.close();
   
   window_height = (int)(screen_height * 0.80 ) - 10;
   window_width =  window_height;
   
   return 0;
}





/**************************************************************************
*
*  clear everything in the window, and then redraw everything 
*  function is typically called when new planes are loaded
* 
*
****************************************************************************/
int redrawWindow(leda_window& winIn, int num)
{
  winIn.clear();
  return refresh(winIn, num);
}



/**************************************************************************
*
*  function to redraw the contents of the graph window
*   also sets the scaling factor, grid style, etc.
*
*  function always returns 0, but could return number corresponding to 
*   error in future revisions
*
****************************************************************************/
int refresh(leda_window& winIn, int option)
{
  
  refreshMessage(winIn, winMsg);
  
  // update circle radius in case window size was modified
  mainCircleRadius = winIn.ymax() * circleScaleFactor;
  
  xOffset = winIn.xmax()/2;
  yOffset = winIn.ymax()/2;

  leda_point tempPt(xOffset, yOffset);
  originPt = tempPt;
   
  winIn.draw_circle(originPt, mainCircleRadius, circleColor);

  switch (int_crosshair) 
    {
    case 0:      //  large cross-hair
      winIn.draw_segment(xOffset, yOffset, xOffset + mainCircleRadius+1, yOffset,  col_largeCrossHair);
      winIn.draw_segment(xOffset, yOffset, xOffset -(mainCircleRadius+1), yOffset, col_largeCrossHair);
      winIn.draw_segment(xOffset, yOffset, xOffset,yOffset + mainCircleRadius+1,   col_largeCrossHair);
      winIn.draw_segment(xOffset, yOffset, xOffset, yOffset-(mainCircleRadius+1),  col_largeCrossHair);
      break;
      
    case 1:      //  small cross-hair
      winIn.draw_point(winIn.xmax()/2, winIn.ymax()/2, col_smallCrossHair);
      break;
      
    case 2:      // no cross-hair
    default: 
      break;
    }
  
  redrawLines(winIn, option);
  
  return 0;

}





/************************************************************************
 *
 * function to redraw the view lines in the "viewport" window
 *  if the second parameter equals 0, then redraw all lines 
 *  if the second parameter <> 0, then draw all lines except that one
 *  if the second parameter == -1, do not redraw anything (for speed)
 *
 ***********************************************************************/
int redrawLines(leda_window &winIn, int option)
{
  leda_circle C(originPt, mainCircleRadius);
  leda_list<leda_point> segPoints;

  // check how many lines to redraw

  if (option == -1)      // don't redraw any lines, exit this function now
    return 0;

  // if very many lines, then just draw enough to fill the viewport
  int drawQuantity = (pA.getPlaneCount() < 1000) ? pA.getPlaneCount() : 1000;

  if (option == 0) {              // none of the lines are highlighted
    for (int i = 1; i <= drawQuantity; i++) {
      segPoints = C.intersection(pA.get_2D_Line(i));
      if (segPoints.size() != 2) {
	winIn.acknowledge("Bad window size!"); 
	return 0;
      }
      winIn.draw_segment(segPoints.front(), segPoints.back());
    }
    return 0;             // done early, so return from here
  }
  
  
  // got this far, so redraw all lines, but do not draw the selected line
  for (int i = 1; i <= drawQuantity; i++)
    {
      segPoints = C.intersection(pA.get_2D_Line(i));
      if (segPoints.size() != 2) {
	winIn.acknowledge("Bad window size!");
	return 0;
      }
      if (i == option)
	;
      else
	winIn.draw_segment(segPoints.front(), segPoints.back(), col_line);
    }

  return 0;
}



/************************************************************************
 *
 * refresh the cursor coordinates in the main window
 *
 ***********************************************************************/
void refreshCursorCoord(leda_window &winIn, double xVal, double yVal)
{

  static leda_string eraseString("");
  leda_color eraseColor(winIn.get_bg_color());

  leda_string str("%+.4f,  %+.4f", 
		  (xVal-xOffset)*xScale/mainCircleRadius,
		  (yVal-yOffset)*yScale/mainCircleRadius);

  // position for the updated mouse coordinates
  double xPos = winIn.xmin()+2;
  double yPos;
  if (winIn.xmax()/winIn.ymax() < 2)
    yPos = winIn.ymax()/winIn.xmax()*4;
  else
    yPos = winIn.xmax()/winIn.ymax();
  
  winIn.draw_text(xPos, yPos, eraseString, eraseColor);
  winIn.draw_text(xPos, yPos, str);

  eraseString = str;    // overwrite this in the background color next time
  return;
}




/***********************************************************************
 *
 *  refresh the user message in the top part of the window
 *  
 *
 ***********************************************************************/
void refreshMessage(leda_window &winIn, const leda_string & message)
{
  double xPos(winIn.xmin() +  1.5);
  double yPos(winIn.ymax() -  0.5);

  // erase old message by overwriting it in the background color
  static leda_string eraseString("");
  leda_color eraseColor(winIn.get_bg_color());

  //  USE THE NEXT 3 LINES IF YOU DON'T WANT TO SHOW MESSAGES 
  //  WHENEVER THE DISPLAY WINDOW IS NOT WIDE ENOUGH
  //  if ((winIn.xmax()-caption.length()) <  message.length())
  //    ;
  //else  { 

  winIn.draw_text(xPos, yPos, eraseString, eraseColor);
  winIn.draw_text(xPos, yPos, message);

  //  USE THE NEXT LINE IF YOU DON'T WANT TO SHOW MESSAGES 
  //  WHENEVER THE DISPLAY WINDOW IS NOT WIDE ENOUGH
  //  }
 
   eraseString = message;               // overwrite this in the background color next time
 
  return;
}
		  




/************************************************************************
 *  function displays a floating "context" menu in the main window
 *   whenever user performs MOUSE-RIGHT-BUTTON click event
 * 
 ************************************************************************/
void popUpMenu(a3_Plane &pA, leda_window &winIn, double winX, double winY)
{
  int choice;
  menu popMenu;

  popMenu.button("Load...");
  popMenu.button("Save...");
  popMenu.button("Graph");
  popMenu.button("Add random");
  popMenu.button("Add default");
  popMenu.button("Delete");
  popMenu.button("Drag");
  popMenu.button("Help");
  popMenu.button("Refresh");
  popMenu.button("Quit");


  if (winX > 85)
    winX = 85;
  
  if (winY < 28)
    winY = 28;


  choice = popMenu.open(winIn, (int)winX*window_width*.01, 
	       (int)(window_height-0.01*winY*window_height));

  switch (choice)
    {
    case 0:           // load a 3D arrangement from a file
      /* 
	 if (pA.hasChanged)
	 winIn.message(saveBeforeDelete);
	 else
      */
      readFromFile(winIn);
      redrawWindow(winIn, 0);
      break;

      
    case 1:           // save the arrangement to a file
      saveToFile(winIn);
      break;
      
    case 2:
      generateOutput(winIn, ledaGraph);
      break;
      
    case 3:           // add a plane with random orientation
      refresh(winIn,0);
      if (pA.getPlaneCount() == maxPlaneCount)
	winIn.acknowledge(noAdd);
      else {
	addRandomPlane(pA, winIn);
	refresh(winIn,0);
      }
      break;

		
    case 4:           // add a plane in default orientation
      refresh(winIn,0);
      if (pA.getPlaneCount() == maxPlaneCount)
	winIn.acknowledge(noAdd);
      else {
	addDefaultPlane(pA, winIn);
	refresh(winIn,0);
      }
      break;
      
    case 5:           // delete lines from the arrangement 
      if (pA.getPlaneCount() == minPlaneCount)
	winIn.acknowledge(noDelete);
      else {
	deletePlane(pA, winIn);
	pA.sortPlanes();
      }
      break;
      
      
    case 6:          // drag planes
      if (pA.getPlaneCount() == 0)
	winIn.acknowledge(noPlanes);
      else
	dragPlanes(pA, winIn);
      break;
      

    case 7:
      HelpFile("Apak3D_Help");
      break;

    case 8:
      refresh(winIn, 0);
      planeQuantity = pA.getPlaneCount();   // refresh SLIDER bar
      winIn.redraw_panel();

      break;    

    case 9:           // user terminates the program
      exit (0);
      break;
      
    default:
      break;
    }
  
  return;

}



/*************************************************************************
 *
 *  function generates random planes equal to the number selected
 *   by the menu SLIDER
 *
 *  function is called initially, and every time user chooses the
 *   'GENERATE RANDOM PLANES' menu item
 *
 *************************************************************************/
void generatePlanes(a3_Plane &pA, leda_window &winIn)
{
  int numPlanes;
  
  if (maxPlaneCount <= sliderMaxPlanes)
    numPlanes = planeQuantity;     // get the number of planes from the slider
  else
    numPlanes = getPlaneQuanFromUser(planeQuantity);


  if (numPlanes == -1)    // user cancelled this operation in the dialog box, so exit early
    return;


   planeQuantity = 0;                 // reset the slider control to zero
   
   winIn.clear();                     // clear all planes from the window
   pA.clear();                        // clear all planes from the arrangement list
  
   winIn.set_cursor(XC_watch);
   
   for (int i = 0; i < numPlanes; i++) {    // generate a new arrangement
     if (addRandomPlane(pA, winIn) != 0) {  // if err, delete everything
       winIn.clear();
       pA.clear();
       break;
     }
   }
   
   winIn.set_cursor(defaultCursor);
   return;
}



/***********************************************************************
 * 
 * function opens dialog box to obtain user input
 * prompts user for integer value, how many random planes to generate
 *
 * if user input is not a valid positive integer, input is ignored
 *
 ***********************************************************************/
int getPlaneQuanFromUser(int defaultNum)
{

  int result = defaultNum;

  // build a dialog box for the user with a default number in it

  leda_panel queryPanel("Generate Planes");
  queryPanel.buttons_per_line(2);
  queryPanel.int_item("Number of random planes: ", result);
  queryPanel.fbutton("&OK");
  queryPanel.button("&Cancel");


  // open the dialog box and get user response
  int userChoice = queryPanel.open(leda_window::center, leda_window::center);
  
  if (userChoice == 0)
    return result;
    
  if (userChoice == 1)
    return -1;
  
}



/*****************************************************************************
 *
 *  function extracts output from the arrangement of planes
 *   
 *  input parameter 'outputType' is an enumerated label defined 
 *   in a3_Plane.h    (e.g.  ledaGraph, gluskapFile, nodeFile);
 *   
 *  the value of 'outputType' is forwarded when calling to_Graph() in the 
 *   arrangement class
 *
 *  thus, function extracts the specified output type from the arrangement
 *   or, if return value from to_Graph() is non-zero, this function 
 *   displays an error message to the user
 *
 *  Also recall: any three planes that intersect to form a point in 
 *   3D space generates a corresponding unique vertex for an undirected graph
 *
 ****************************************************************************/
void generateOutput(leda_window &winIn, const int outputType)
{

  int status, plane1, plane2, plane3;    // indices for bad planes if graph fails
  
  winIn.set_cursor(XC_watch);

  leda_string fileName;

  ofstream outFile;

  if (outputType != ledaGraph)
    {
      int result = getFileNameFromUser(winIn, fileName);

      while (result >= 0) {
	if (result == 1) {
	  if (outputType == gluskapFile)
	    HelpFile("SaveGL_Help");
	  else
	    HelpFile("SaveNL_Help");
	  result = getFileNameFromUser(winIn, fileName);
	}
	
	else {
	  outFile.open(fileName);
	  if (! outFile.is_open()) {
	    winIn.acknowledge(ofileOpenErr);
	    result = getFileNameFromUser(winIn, fileName);
	  }
	  else
	    break;
	}
      }
      if (result < 0)
	return;
    }
  
  // got this far so outFile path is valid and file has write permissions
  // close the output file for now
  outFile.close();
  delete_file(fileName);

  winIn.disable_panel();
  status = pA.to_Graph(outputType, fileName, plane1, plane2, plane3); 
  winIn.enable_panel();

  leda_string planeMsg("Graph failed while processing planes %i, %i and %i. ", plane1, plane2, plane3);
  winMsg = blankMsg;
  
  switch (status)
    {
    case 0:    // no errors found
      break;
      
    case 1:    // degenerate arrangement, found two coincident planes
      winMsg = "Found coincident planes.";
      planeMsg += winMsg;
      break;

    case 2:    // degenerate arrangement, found two parallel planes
      winMsg = "Found parallel planes.";
      planeMsg += winMsg;
      break;

    case 3:    // degenerate arrangement, 3 planes intersection yields a line
       winMsg = "3-plane intersection returned a line.";
      planeMsg += winMsg;
      break;

    case 4:    // degenerate arrangement, two vertices located at same 3D coordinate
      winMsg = "Found two graph nodes at same 3D coordinate.";
      planeMsg += winMsg;
      break;

   case 5:     // incorrect number of nodes due to unspecified error 
      planeMsg = "Graph failed. Incomplete node set.";
      break;
      
    case 6:    // bad argument in call to function
      planeMsg = "Incorrect output type specified. No file generated.";
      break;

    default:
      winMsg = "Non-specified error.";
      planeMsg += winMsg; 
      break;
    }
  
  winIn.set_cursor(defaultCursor);
  winMsg = blankMsg;     // reset this to default

  if (status != 0) {      // if graph failed, display error message to user
    refresh(winIn,0);
    refreshMessage(winIn, planeMsg);  
    int winEvent, eventVal;  
    double xEvent, yEvent;   
    
    // keep the error message in the display until user 
    // performs a click with a mouse button
    winEvent = winIn.read_event(eventVal, xEvent, yEvent);
    while (winEvent != button_press_event) {
      switch (winEvent) {
      case key_press_event:        // do nothing for keyboard action
      case key_release_event:
	break;
	
      case configure_event:        // if user moves the window, refresh it
      case panel_press_event:
      case exposure_event:
	refresh(winIn, 0);
	refreshMessage(winIn, planeMsg);
	break;
	
      case motion_event:   // if flag is set, update cursor coordinates
	if (bool_showCoord)
	  refreshCursorCoord(winIn, xEvent, yEvent);    
	break;
	
      default:
	break;
      }
      winEvent = winIn.read_event(eventVal, xEvent, yEvent);
    }  
    
    put_back_event();    // last event was a BUTTON_PRESS or MOUSE_CLICK
                         // so let the main user loop handle it
  }

  return;
}




/***************************************************************************
 *
 * function adds a random plane to the arrangement
 * 
 * pre:  number of planes < maxPlanecount
 *
 *  if call to a3_Plane::add_Random_Plane() returns error code
 *   then function does nothing
 * 
 * post: add'l plane is added 
 *       if totalPlanes = maxPlaneCount then
 *        send error message to user and do not add a plane
 *       
 *
 **************************************************************************/
int addRandomPlane(a3_Plane &pA, leda_window &winIn)
{
  planeQuantity = pA.getPlaneCount();     // ensure SLIDER is set to correct number
  winIn.redraw_panel();

  if (pA.add_Random_Plane() == -1){      // error, so terminate operation
       winIn.acknowledge(addFailMsg);
       return -1;
     }

   else {
      planeQuantity ++;       // update the SLIDER on the menu panel
      winIn.redraw_panel();
   }

   return 0;
}



/*****************************************************************************
 *
 * function adds a plane to the window
 *
 *  default orientation of the new plane is x-y
 *   (i.e.  z=0 for all points in the new plane)
 *
 *  if function call to add_Default_Plane failed, nothing is added;
 *  if totalPlanes = maxPlaneCount then send error message to user
 *    and do not add plane
 *
 ****************************************************************************/
void addDefaultPlane(a3_Plane &pA, leda_window &winIn)
{
  planeQuantity = pA.getPlaneCount();      // ensure SLIDER is set to correct number
  winIn.redraw_panel();
  
  if (pA.add_Default_Plane() == -1)
    winIn.acknowledge(addFailMsg);
  else {
    planeQuantity ++;       // update the SLIDER on the menu panel
    winIn.redraw_panel();
  }
  return;
}



/*****************************************************************************
 *
 * function removes a plane from the window
 *
 * pre:  number of planes > 0
 *
 *  if call to a3_Plane::delete_Plane() returns error code
 *   then function sends error message to user
 * 
 * post: selected plane is deleted 
 *       if total planes = 0 then return from this function
 *       
 *
 ****************************************************************************/
void deletePlane(a3_Plane &pA, leda_window &winIn)
{
   int buttonVal = 0;
   double x, y;
   winMsg = delPrompt;
   winIn.set_cursor(XC_hand2);
   
   planeQuantity = pA.getPlaneCount();       // ensure SLIDER is set to correct value
   winIn.redraw_panel();

   refresh(winIn,0);
   winIn.disable_panel();

   // get x-y coordinates close to the line/plane to delete
   buttonVal = winIn.read_mouse(x, y);

   while (buttonVal != MOUSE_BUTTON(3))
     {
       if (buttonVal == MOUSE_BUTTON(1)) {
	 if (pA.delete_Plane(x, y) == -1)
	   winIn.acknowledge(delFailMsg);
	 else {
	   planeQuantity--;          // update the SLIDER on the menu panel
	   winIn.redraw_panel();     // refresh the SLIDER position
	 }
       }
       redrawWindow(winIn,0);

       if (pA.getPlaneCount() == minPlaneCount)      // lower limit for deletion
	 break;                                      // so quit the loop
       
       buttonVal = winIn.read_mouse(x, y);
     }

   // cleanup the window
   winMsg = blankMsg;
   winIn.set_cursor(defaultCursor);
   refresh(winIn,0);
   winIn.enable_panel();
   return;
}





/****************************************************************************
 *
 *  set the working directory for file I/O operations (File-Load, File-Save)
 * 
 *  function opens a dialog box for the user to enter a valid pathname 
 *  
 *  if pathname exists but user does not have read/write permissions, 
 *   generate an error message to the user  
 *
 *  if pathname does not exist, generate an error message to the user
 *
 *  if pathname is valid and has read/write permissions, set the working path
 *   accordingly
 *
 *
 ***************************************************************************/
void setPath(leda_window &winIn)
{
  
  bool pathIsGood = false;
  leda_string newDir(filePath);

  leda_panel pathPanel("Set Working Directory");
  pathPanel.buttons_per_line(5);
  pathPanel.string_item("Path: ", newDir);
  pathPanel.fbutton("&Set Path");
  pathPanel.button("&Help");
  pathPanel.button("&Cancel");

  
  for (;;) {
      int userChoice = pathPanel.open(leda_window::center, leda_window::center);
      
      if (userChoice == 0) {

	if (newDir.tail(1) != "/")        // check for missing slash
	  newDir += "/";
	
	if (! is_directory(newDir)) {    // bad pathname, so return an error
	  winIn.acknowledge(badPathName);
	  continue;
	}
     	
	else {          // make sure we have READ and WRITE permissions on the directory
	  ofstream outFile;

	  leda_string tmpFile(tmp_file_name());       // generate a unique temp file name

	  // for LEDA 4.2 UNIX, strip off the first 9 char
	  tmpFile = tmpFile(9, tmpFile.length()-1);   // get rid of LEDA's path prefix

	  // for LEDA 4.1 LINUX, strip off the first 5 char
	  //  tmpFile = tmpFile(5, tmpFile.length()-1);   // get rid of LEDA's path prefix
	  
	  tmpFile = newDir + tmpFile;             // try read/write to the new pathname

	  outFile.open(tmpFile);
	  if (!outFile.is_open())
	    winIn.acknowledge(pathErr);
	  else {
	    outFile << "ArrangePak3D Test File - Delete Me ";
	    outFile.close();
	    ifstream inFile;
	    inFile.open(tmpFile);
	    if (!inFile)
	      winIn.acknowledge(pathErr);
	    else
	      pathIsGood = true;
	    delete_file(tmpFile);                      // get rid of the temp file
	  }
	  
	}
	if (pathIsGood) {
	  filePath = newDir;
	  break;
	}
      }
  
      if (userChoice == 1) {
	HelpFile("PathHelp");
	continue;
      }
      
      if (userChoice = 2)
	break;
  }
 
  return;
}    





/****************************************************************************
 *
 *  prompt the user for a file name
 *  
 *  function will loop until user gives valid string data
 *   or cancels operation
 *
 *  returns zero if user provides valid string
 *   otherwise returns non-zero
 *
 ***************************************************************************/
int getFileNameFromUser(leda_window &winIn, leda_string &userFile)
{
  winIn.set_cursor(defaultCursor);

  int functionResult = 0;
  
  leda_string currDir(filePath);
  leda_string fileName;
     
  if (currDir.tail(1) != "/")        // check for missing slash
    currDir += "/";
  
  leda_panel savePan("Save File");
  savePan.buttons_per_line(5);
  savePan.string_item("Filename: ", fileName);
  savePan.fbutton("&Save");
  savePan.button("&Help");
  savePan.button("&Cancel");

  
  for (;;) {
      int userChoice = savePan.open(leda_window::center, leda_window::center);
      
      if (userChoice == 0) {
	if (fileName == "") {       // missing filename, so return an error
	  winIn.acknowledge(fileNameErr);
	  continue;
	}
	else {
	  userFile = currDir + fileName;
	  break;
	}
      }
      
      if (userChoice == 1) {   // user chooses HELP button
	functionResult = 1;
	break;
      }
      
      if (userChoice = 2) {    // user chooses CANCEL button
	functionResult = -1;
	break;
      }
  }
      
  return functionResult;

}




/****************************************************************************
 *
 *  save list of planes to file 
 *  output format is POINT-VECTOR notation
 *  first value is a point contained in the plane as (x,y,z)
 *   followed by a '3' which signifies a vector in 3-dimensions
 *    followed by 3 doubles, comprising a vector normal to the plane
 *
 *  e.g.    (10,-20,30)  3  0 1.5 3
 *
 *  output file begins with a six line header
 *
 ***************************************************************************/
void saveToFile(leda_window &winIn)
{
  if (pA.getPlaneCount() == 0) {
    winIn.message(noPlanes);
    return;
  }

  //  leda_string currDir(filePath);

  leda_string fileName;
  ofstream outFile;

  int result = getFileNameFromUser(winIn, fileName);

  while (result >= 0) {
    if (result == 1) {
      HelpFile("SaveHelp");
      result = getFileNameFromUser(winIn, fileName);
    }      
    else {
      outFile.open(fileName);
      if (! outFile.is_open()) {
	winIn.acknowledge(ofileOpenErr);
	result = getFileNameFromUser(winIn, fileName);
      }
      else
	break;
    }
  }

  if (result < 0)
    return;
  
  if (! pA.savePlanesToFile(outFile))
    winIn.acknowledge(writeErr);
  else 
    winIn.message(fileSaved);
  
  return;
}    



/****************************************************************************
 *
 *  read list of planes from file 
 *  input format is POINT-VECTOR notation
 *  first value expected is a leda_d3_point 
 *  second value expected is a leda_vector in 3-dimensions
 *  
 *  e.g.    (10,-20,30)  3  0 1.5 3
 *
 *  reader ignores six-line header at top of file
 *
 ***************************************************************************/
void readFromFile(leda_window &winIn)
{
  
  leda_string currDir(filePath);
  leda_string fileName;
  
  if (currDir.tail(1) != "/")        // check for missing slash
    currDir += "/";

  ifstream inFile;
  
  leda_panel openPan("Open File");
  openPan.buttons_per_line(5);
  openPan.string_item("Filename: ", fileName);
  openPan.fbutton("&Open");
  openPan.button("&Help");
  openPan.button("&Cancel");

  for (;;) {
    int userChoice = openPan.open(leda_window::center, leda_window::center);
    
    if (userChoice == 0) {
      if (fileName == "") {       // missing filename, so return an error
	winIn.acknowledge(fileNameErr);
	continue;
      }
      
      inFile.open(currDir + fileName);
      if (! inFile.is_open()) {            // problem with the given filename
	winIn.acknowledge(ifileOpenErr);
	continue;
      }
      
      if (! pA.readPlanesFromFile(inFile)) {
	planeQuantity = 0;
	winIn.acknowledge(readErr);
	refresh(winIn,0);
      }
      else {
	openPan.close();
	planeQuantity = pA.getPlaneCount();
	winIn.redraw_panel();
	refresh(winIn,0);
	break;
      }
    }
    
    if (userChoice == 1) {     // display the help file
      HelpFile("LoadHelp");
      continue;
    }
    
    if (userChoice = 2) {
      openPan.close();
      break;
    }
  }
  
  return;
  
}
  


/*****************************************************************************
 *
 * function manipulates planes in the arrangement
 * 
 ******************************************************************************/
void dragPlanes(a3_Plane &pA, leda_window &winIn)
{
  int buttonVal;                // variable to store the mouse button that is clicked
  int planeNum;                 // index into the list of planes, correspond to chosen plane
  leda_window *wPtr = NULL;     // pointer to window that mouse event occurs within
  int winNumber = 0;            // enumerated value corresponding to the window of the event
  int winEvent = 0;             // variable to store the window event (click, motion, re-size, etc)
  int eventValue = 0;           // value of the window event (eg. which button or key was pressed) 
  double x, y;                  // mouse coordinates at the time of the event
  bool doneFunc = false;
  bool doneDrag = false;
  bool protMode;                // protection mode prevents rotating end over end - for future use
  int grabIndex;

  double xOld, yOld, xNew, yNew;
  leda_point oldPosition, newPosition;
  leda_segment frontEdge;
  leda_d3_point newPt(0,0,0);
  int pointNum = 0;
  leda_point grabPoint;
  double grabDist;
  leda_d3_segment rotateAxis;
  leda_segment midLine;
  leda_d3_plane grabbedPlane;


  planeQuantity = pA.getPlaneCount();   // ensure SLIDER is set to correct number
  winIn.redraw_panel();

  // place a user prompt in the main window
  winMsg = choosePlane; 
  winIn.set_cursor(XC_hand2);
  refresh(winIn,0);
  winIn.disable_panel();

  do {
    winEvent = read_event(wPtr, eventValue, x, y);
    if (wPtr == &winIn) {
      switch (winEvent) {
	
      case configure_event:
      case panel_press_event:
      case exposure_event:
	refresh(winIn, 0);
	break;
	
      case button_press_event:
	buttonVal = eventValue;
	switch(buttonVal) {
	case MOUSE_BUTTON(3):                      // RIGHT-CLICK, so exit 
	  winMsg = blankMsg;
	  winIn.set_cursor(defaultCursor);
	  refresh(winIn, 0);
	  doneFunc = true;
	  break;
	  
	case MOUSE_BUTTON(1):                      // LEFT-CLICK, so get data on closest plane item
	  doneDrag = false; 
	  protMode = false;
	  planeNum = pA.findPlane(x, y);           // find out which plane is closest to the click
	  pA.getPlane(planeNum, grabbedPlane);     // make a local copy of the plane
	  midLine = get_2D_Axis(grabbedPlane);     // make a local copy of the intersecting segment in viewPort
	  
	  winMsg = dragPlane;                      // post a message to user
	  winIn.set_cursor(defaultCursor);         
	  redrawWindow(winIn, planeNum);           // draw all lines in background color, except selected plane

	  drawPlane(winIn, grabbedPlane, frontEdge, planeNum);  // draw selected plane in pseudo-3D perspective and
                                                                // get a copy of the frontEdge by reference


	  // stay in this loop until user does a RIGHT-CLICK on mouse
	  do {
	    winEvent = read_event(wPtr, eventValue, x, y);    // wait for next mouse event
	    if (wPtr == &winIn) {
	      leda_point oldPos(x,y);	      
	      switch (winEvent) {
	      
	      case configure_event:
	      case panel_press_event:
	      case exposure_event:
		refresh(winIn, planeNum);
		drawPlane(winIn, grabbedPlane, frontEdge, planeNum);
		break;
		
	      case button_press_event:
		buttonVal = eventValue;
		
		switch(buttonVal) {  

		case MOUSE_BUTTON(1):
		  winEvent = read_event(wPtr, eventValue, x, y);
		
		  getGrabPoint(grabbedPlane, frontEdge, midLine, oldPos,     // input parameters
			       grabPoint, rotateAxis, grabDist, grabIndex, protMode);   // output by reference
		    
		  if (grabDist < 3.0) {
		    while (winEvent == motion_event) {  //  WHILE LEFT MOUSE IS DOWN
		      
		      xNew = x; 
		      yNew = y;
		      leda_point newMousePoint(xNew, yNew);
		      leda_segment dragSeg(grabPoint, newMousePoint);
		      

		      winIn.start_buffering();
		      // winIn.draw_segment(dragSeg, "red");

		      leda_segment rotAxis(rotateAxis.project_xy());
		      leda_line roAxis(rotAxis.start(), rotAxis.end());
		      winIn.draw_line(roAxis, "green");

		      if (pA.getPlaneCount() < 100)
			redrawWindow(winIn, planeNum);         // redraw everything
		      else
			redrawWindow(winIn, -1);               // for speed, don't redraw background lines
		      
		      manipulate_3D_Plane(winIn, grabbedPlane, newMousePoint, rotateAxis, grabIndex, protMode);
		      drawPlane(winIn, grabbedPlane, frontEdge, planeNum);

                      if (bool_showCoord)
                        refreshCursorCoord(winIn, xNew, yNew);
                      
		      winIn.stop_buffering();
		      winIn.flush_buffer();

     		      xOld = x;
     		      yOld = y;

		      winEvent = read_event(wPtr, eventValue, x,y);
		      winIn.draw_segment(dragSeg, "white");
		      
		    }   // while (winEvent == motion_event)   // while LEFT MOUSE_BUTTON DOWN
		  
		    // insert the modified plane back into the list, replacing the original one
		    pA.swapPlane(grabbedPlane, planeNum);

		    pA.getPlane(planeNum, grabbedPlane);     // update local copy of the plane
		    midLine = get_2D_Axis(grabbedPlane);     // update local copy of axis

		  }     // if (grabDist < 3)
		  break;
		  
		case MOUSE_BUTTON(3):	
		  winMsg = choosePlane;
		  winIn.set_cursor(XC_hand2);
		  redrawWindow(winIn, 0);
		  doneDrag = true;
		  break;
		  
		default:
		  redrawWindow(winIn, 0);
		  break;
		}
		break;      // case: button_press_event
		
		
	      default:
		break;
		
	      }   // switch(winEvent)
	     
	    }  // if (wPtr == &winIn)  

	  } while (!doneDrag);
	  
	} // switch(buttonVal)
	
      }  // switch(winEvent)
      
    } // if (wPtr == &winIn)
    
    // if wPtr == some other window or if a motion event
    // do something else here
    
   } while (!doneFunc);

  // clean up window
  winMsg = blankMsg;
  refresh(winIn, 0);
  winIn.enable_panel();
  return;
}
   


/******************************************************************
 *  function draws a "pseudo-3D" representation of a plane in the 
 *   viewport
 *  also draws the unit vector coordinates for the plane in the
 *  Upper Right Hand Corner of the window
 *
 ******************************************************************/
leda_segment drawPlane(leda_window &winIn, leda_d3_plane &planeIn, 
		       leda_segment &frontEdge, const int& planeNum)
{
  double xAngle, yAngle, zAngle;
  leda_circle C(originPt, mainCircleRadius);
  leda_list<leda_point> segPoints;
  leda_d3_point p(xOffset, yOffset, 0);
  leda_d3_point q, r;
  leda_vector u,v;

  winIn.set_node_width(mainCircleRadius/14.0);
  v = planeIn.normal();
  u = v.norm();

  zAngle = sqrt(pow(u.zcoord(), 2));     // get the absolute value of the z-coord

  // ******* parabolic function defines "damping" or "response"	 
  //  zVal = pow(cos(zAngle*1.57),2)*.1+10;

  //****  parabolic function version 2
  //  zVal = 2*pow(zAngle,2);  

  // try a linear function, essentially an inverted V
  // when z-coordinate of the vector normal to the
  // plane is close to zero
  /*
  if (zAngle < 0.0005)
    zVal = -1500*zAngle + 1500;

  else if (zAngle < 0.001)
    zVal = -1000*zAngle + 1000;
  */
 
  // logarithmic approach, not very smooth
  // zVal = pow(5.0, fabs(log10(zAngle)));

  // final correction factor
  //  if (zVal < 1) zVal = 1;

  // diagnostic message
  //  cerr << "      zAngle/zVal " << zAngle << " / " << zVal << endl;
 
  int rotDir = (u.xcoord() > 0) ? 1 : -1;
  
  leda_d3_point fPt1(0.0, 0.0, zVal);
  leda_d3_point fPt2(1.0, 0.0, zVal);
  leda_d3_point fPt3(1.0, 1.0, zVal);
  leda_d3_plane newFront(fPt1, fPt2, fPt3);
  frontFace = newFront;


  // make a vertical plane from 3 non-collinear points in 3D space at (x,y,0)
  leda_d3_point pt1(0.0, 0.0, 0.0);
  leda_d3_point pt2(1.0, 0.0, 0.0);
  leda_d3_point pt3(1.0, 1.0, 0.0);
  leda_d3_plane viewFace(pt1, pt2, pt3);

  viewFace.intersection(planeIn, q, r);    // q,r define points of intersection between planeIn and viewFace
  leda_d3_line viewLine3D(q,r);            // viewLine3D is 3D line of intersection between planeIn and (x,y,0)

  leda_line viewLine;                      
  viewLine3D.project_xy(viewLine);         // viewLine is the 2D projection of viewLine3D onto (x,y)

  segPoints = C.intersection(viewLine);
  if (segPoints.size() != 2)
    {
      winIn.acknowledge("a3d::drawPlane list size is not 2. Notify vendor.");
      cerr << "BIG ERROR !! " << viewLine;
      exit (-1);
    }
  
  leda_segment centerAxis(segPoints.front(), segPoints.back());


  if (frontFace.intersection(planeIn, q, r) == 1)     // same as above, but for planeIn and frontFace
    {
      leda_segment seg2d;               
      leda_d3_line line1(q,r);
      leda_line line2;                // line2 will be the 2D line of intersection between plane and circle
      line1.project_xy(line2);     
      seg2d = line2.seg();            
      
      leda_point midPt = midpoint(segPoints.front(), segPoints.back());
      leda_segment segCtr(segPoints.front(), segPoints.back());

      winIn.draw_filled_node(segPoints.front(), "red");
      winIn.draw_filled_node(segPoints.back(), "red");


      leda_segment segDir(seg2d.perpendicular(midPt));
      double D = segDir.direction();
      double revD = segDir.reverse().direction();

      double lineGap = 8.0*pow(sin(zAngle), (1.0/3.0));
      
      double extendLength;
      if (zAngle > 0.54)
	extendLength = 4*pow(acos(zAngle), 2.0); 
      else
	extendLength = 2*pow(acos(zAngle), 1.5);

      leda_segment segA(segCtr.start(), D, lineGap);
      leda_segment segB(segCtr.end(), D, lineGap);

      // construct a segment corresponding to the front edge
      leda_segment frontSeg(segA.end(), segB.end());
      leda_segment backSeg(frontSeg.reflect(midPt));

      // extend the front edge segment - make it appear closer to the user
      leda_segment stretch1(frontSeg.start(), frontSeg.direction(), 
			    frontSeg.length()+ extendLength);

      leda_segment stretch2(stretch1.end(), (stretch1.reverse()).direction(),
			    stretch1.length()+ extendLength);

      leda_segment shrink1(backSeg.start(), backSeg.direction(), backSeg.length() - extendLength);
      leda_segment shrink2(shrink1.end(), (shrink1.reverse()).direction(), shrink1.length()-extendLength);


      // draw the 3D plane
      winIn.draw_line(segCtr, col_rotateAxis);
      winIn.draw_segment(segCtr, col_pivotLine);
      winIn.draw_segment(shrink2, col_backEdge);
      winIn.draw_segment(stretch2, col_frontEdge);

      if (stretch2.start().cmp_dist(shrink2.start(), shrink2.end()) == 1) {
	winIn.draw_segment(stretch2.start(), shrink2.end(), col_sideEdge);
	winIn.draw_segment(stretch2.end(), shrink2.start(), col_sideEdge);
      }
      else  {
	winIn.draw_segment(stretch2.start(), shrink2.start(), col_sideEdge);
	winIn.draw_segment(stretch2.end(), shrink2.end(), col_sideEdge);
      }

      winIn.draw_filled_node(stretch2.start(), "green");
      winIn.draw_filled_node(stretch2.end(), "orange");
      winIn.draw_filled_node(midpoint(stretch2.start(), stretch2.end()), "yellow");

      frontEdge = stretch2;
    }

  // draw the coordinates for the Unit Vector normal to the plane from the origin
  leda_string unitVec_x("x:  % .6f", u.xcoord());
  leda_string unitVec_y("y:  % .6f", u.ycoord());
  leda_string unitVec_z("z:  % .6f", u.zcoord());
  leda_string planeIndexString("Plane # %i  ", planeNum);

  
//  double x_stringPos = winIn.xmax() - caption.length()+4;
  double x_stringPos = winIn.xmax()-caption.length();
  double y_stringPos;
  if (winIn.xmax()/winIn.ymax() < 2)
    y_stringPos = winIn.ymax() - winIn.ymax()/winIn.xmax();
  else
     y_stringPos = winIn.ymax() - winIn.xmax()/winIn.ymax();
  /* IF _DEF_LINUX_
     double x_coordPos = winIn.xmax() - unitVec_x.length()+1;
  */
  /* IF _DEF_UNIX_ */
  double x_coordPos = winIn.xmax() - unitVec_x.length()-3;
 
  winIn.draw_text(x_stringPos, y_stringPos, caption);
  winIn.draw_text(x_coordPos, y_stringPos-2, unitVec_x);
  winIn.draw_text(x_coordPos, y_stringPos-4, unitVec_y);
  winIn.draw_text(x_coordPos, y_stringPos-6, unitVec_z);
  winIn.draw_text(x_coordPos, y_stringPos-8, planeIndexString);

  return centerAxis;

}





/********************************************************************
 *
 *  shows all planes in "pseudo-3D" representation in the viewport
 *
 *
 *******************************************************************/
void showAllPlanes(leda_window & winIn)
{
  double xAngle, yAngle, zAngle;
  leda_circle C(originPt, mainCircleRadius);
  leda_list<leda_point> segPoints;
  leda_d3_point p(xOffset, yOffset, 0);
  leda_d3_point q, r;
  leda_vector u,v;
  leda_d3_plane pickPlane;

  winIn.set_node_width(mainCircleRadius/14.0);

  winMsg = blankMsg;
  redrawWindow(winIn, 0);

  planeQuantity = pA.getPlaneCount();
  winIn.redraw_panel();

  if (pA.getPlaneCount() > 0)
    {
      winIn.disable_panel();
      for (int p = 1; p <= pA.getPlaneCount(); p++)
	{
	  pA.getPlane(p, pickPlane);
	  v = pickPlane.normal();
	  u = v.norm();
	  
	  zAngle = sqrt(pow(u.zcoord(), 2));     // get the absolute value of the z-coord
	  
	  leda_d3_point fPt1(0.0, 0.0, zVal);
	  leda_d3_point fPt2(1.0, 0.0, zVal);
	  leda_d3_point fPt3(1.0, 1.0, zVal);
	  leda_d3_plane newFront(fPt1, fPt2, fPt3);
	  frontFace = newFront;
	  
	  // make a vertical plane from 3 non-collinear points in 3D space at (x,y,0)
	  leda_d3_point pt1(0.0, 0.0, 0.0);
	  leda_d3_point pt2(1.0, 0.0, 0.0);
	  leda_d3_point pt3(1.0, 1.0, 0.0);
	  leda_d3_plane viewFace(pt1, pt2, pt3);
	  
	  viewFace.intersection(pickPlane, q, r);   // q,r define pts of intersection between plane & viewFace
	  leda_d3_line viewLine3D(q,r);             // viewLine3D is line of intersection between plane & (x,y,0)
	  
	  leda_line viewLine;                      
	  viewLine3D.project_xy(viewLine);         // viewLine is the 2D projection of viewLine3D onto (x,y)
	  
	  segPoints = C.intersection(viewLine);
	  if (segPoints.size() != 2)
	    {
	      winIn.acknowledge("a3d::drawPlane list size is not 2. Notify vendor.");
	      cerr << "BIG ERROR !! " << viewLine;
	      exit (-1);
	    }
	  
	  leda_segment centerAxis(segPoints.front(), segPoints.back());
	  
	  if (frontFace.intersection(pickPlane, q, r) == 1)     // same as above, but for plane and frontFace
	    {
	      leda_segment seg2d;               
	      leda_d3_line line1(q,r);
	      leda_line line2;                // line2 will be the line of intersection between plane and circle
	      line1.project_xy(line2);     
	      seg2d = line2.seg();            
	      
	      leda_point midPt = midpoint(segPoints.front(), segPoints.back());
	      leda_segment segCtr(segPoints.front(), segPoints.back());
	

	      leda_segment segDir(seg2d.perpendicular(midPt));
	      double D = segDir.direction();
	      double revD = segDir.reverse().direction();

	      double lineGap = 8.0*pow(sin(zAngle), (1.0/3.0));
          
	      double extendLength;
	      if (zAngle > 0.54)
		extendLength = 4*pow(acos(zAngle), 2.0); 
	      else
		extendLength = 2*pow(acos(zAngle), 1.5);
	   
	      leda_segment segA(segCtr.start(), D, lineGap);
	      leda_segment segB(segCtr.end(), D, lineGap);

	      // construct a segment corresponding to the front edge
	      leda_segment frontSeg(segA.end(), segB.end());
	      leda_segment backSeg(frontSeg.reflect(midPt));
	      
	      // extend the front edge segment - make it appear closer to the user
	      leda_segment stretch1(frontSeg.start(), frontSeg.direction(), 
				    frontSeg.length()+ extendLength);
	      
	      leda_segment stretch2(stretch1.end(), (stretch1.reverse()).direction(),
				    stretch1.length()+ extendLength);
	      
	      leda_segment shrink1(backSeg.start(), backSeg.direction(), backSeg.length() - extendLength);
	      leda_segment shrink2(shrink1.end(), (shrink1.reverse()).direction(), shrink1.length()-extendLength);
	      
	      
	      // draw the 3D plane
	      winIn.draw_segment(segCtr, col_rotateAxis);
	      winIn.draw_segment(segCtr, col_pivotLine);
	      winIn.draw_segment(shrink2, col_backEdge);
	      winIn.draw_segment(stretch2, col_frontEdge);
	      
	      if (stretch2.start().cmp_dist(shrink2.start(), shrink2.end()) == 1) {
		winIn.draw_segment(stretch2.start(), shrink2.end(), col_sideEdge);
		winIn.draw_segment(stretch2.end(), shrink2.start(), col_sideEdge);
	      }
	      else  {
		winIn.draw_segment(stretch2.start(), shrink2.start(), col_sideEdge);
		winIn.draw_segment(stretch2.end(), shrink2.end(), col_sideEdge);
	      }
	      
	    }
	}
    }
  
  leda_string quitMsg("Right-click anywhere in the window to resume.");
  refreshMessage(winIn, quitMsg);


  // wait for a right mouse click before resetting to 2-dimensional view
  bool doneLoop = false;
  int winEvent, eventVal;
  double xEvent, yEvent;     // disregard these values, but need the variables

  while (! doneLoop) {
    winEvent = winIn.read_event(eventVal, xEvent, yEvent);
    
    switch (winEvent) {
    case key_press_event:        // do nothing for keyboard action
    case key_release_event:
    case configure_event:        // if user moves the window, do nothing
    case panel_press_event:      // this might allow the window to go blank
    case exposure_event:
    case motion_event:
      break;
	
    case button_press_event:
      if (eventVal == MOUSE_BUTTON(3))
	doneLoop = true;
      break;
      
    default:
      break;
    }
  }
  refresh(winIn, 0);
  winIn.enable_panel();    
  return;
  
}




/***************************************************************
 *
 *  function modifies highlighted plane during  user-interactive
 *     plane transforms 
 *
 *  finds a point in 3D space corresponding to the location 
 *    point of the user's mouse
 * 
 *  'rotateAxis' dictates the axis around which the plane is  
 *    rotated
 *
 *  modifiedPlane is the result of dragging the 'grabbedPlane'
 *   around the 'rotateAxis' to the 'newMousePoint'
 *
 *  function returns the resulting transformation by reference
*    variable called 'grabbedPlane'
 **************************************************************/
void manipulate_3D_Plane(leda_window &winIn, leda_d3_plane &grabbedPlane, 
			 leda_point &newMousePoint, leda_d3_segment &rotateAxis, 
			 const int &grabIndex, const bool &protMode)
{
  leda_d3_point dragPoint;

  if (grabIndex == 3 || grabIndex == 4)
    dragPoint = convertTo3D(newMousePoint, 0);
  else
    dragPoint = convertTo3D(newMousePoint, 1);

  leda_d3_plane modifiedPlane(dragPoint, rotateAxis.start(), rotateAxis.end());
  grabbedPlane = modifiedPlane;

  return;
}


/***************************************************************
 *
 *  returns a segment in 2D space corresponding to that portion
 *   of a plane which intersects the viewport circle
 *
 *  the 3D plane under test is given by an index into the list 
 *
 *
 **************************************************************/
leda_segment get_2D_Axis(const leda_d3_plane& thePlane)
{
  leda_d3_point q, r;
  leda_circle C(originPt, mainCircleRadius);
  leda_list<leda_point> segPoints;

  leda_segment centerAxis;       // corresponds to the line where plane intersects viewport, projected onto xy

  // make a vertical plane from 3 non-collinear points in 3D space at (x,y,0)
  leda_d3_point pt1(0.0, 0.0, 0.0);
  leda_d3_point pt2(1.0, 0.0, 0.0);
  leda_d3_point pt3(1.0, 1.0, 0.0);
  leda_d3_plane viewFace(pt1, pt2, pt3);

  viewFace.intersection(thePlane, q, r);   // q,r define points of intersection between plane and viewFace
  leda_d3_line viewLine3D(q,r);            // viewLine3D is 3D line of intersection between planeIn and (x,y,0)

  leda_line viewLine;                      
  viewLine3D.project_xy(viewLine);         // viewLine is the 2D projection of viewLine3D onto (x,y)

  segPoints = C.intersection(viewLine);
  if (segPoints.size() != 2)
    {
      cerr << "a3d::get_2D_Axis - list size is not 2. Notify vendor.\n";
      exit (-1);
    }
  
  leda_segment d2_result(segPoints.front(), segPoints.back());
 
  return d2_result;
}






/***********************************************************
 *
 *  calculate whether mouse click was close to the front
 *   edge, corner point or end point of a 3D plane in the
 *   viewport
 *
 *  function is called by dragPlanes
 *
 *
 *  possible grabIndex =  { 1,2,3,4,5}
 *  grabPoints map to -->  { corner1, corner2, end1, end2, frontEdge }
 *
 *
 **********************************************************/
void getGrabPoint(const leda_d3_plane &grabbedPlane,
		  const leda_segment &frontEdge,
		  leda_segment &midLine, leda_point &mousePoint,
		  leda_point &grabPoint, leda_d3_segment &rotateAxis,
		  double &grabDist, int &grabIndex, bool &protMode)
{

  grabIndex = 0;         // default to grabbing front edge
  double cDist, cornerDist1, cornerDist2; 
  double eDist, endDist1, endDist2; 
  double edgeDist;
  

  // find the midpoint of the 2-dimensional axis of rotation, and convert to 3D
  leda_point midPoint_2d(midpoint(midLine.start(), midLine.end()));

  leda_d3_point midPoint_3d(convertTo3D(midPoint_2d, 0));

  leda_d3_point rotAxis1, rotAxis2;
  
  leda_segment mouseSeg;

  int grabCorner, grabEnd;

  // orient the midLine axis and frontEdge segment the same way
  if (frontEdge.start().distance(midLine.end()) < frontEdge.start().distance(midLine.start()))
    midLine = midLine.reverse();
  
  // calculate whether the user grabbed a corner, side, or front edge
  cornerDist1 = frontEdge.start().distance(mousePoint);
  cornerDist2 = frontEdge.end().distance(mousePoint);
  endDist1 = midLine.start().distance(mousePoint);
  endDist2 = midLine.end().distance(mousePoint);
  edgeDist = frontEdge.distance(mousePoint);

  if (cornerDist1 > cornerDist2) {
    grabCorner = 2;
    cDist = cornerDist2;
  }
  else {
    grabCorner = 1;
    cDist = cornerDist1;
  }

  
  if (endDist1 > endDist2) {
    grabEnd = 2;
    eDist = endDist2;
  }
  else {
    grabEnd = 1;
    eDist = endDist1;
  }
 
  if (eDist < cDist && eDist < edgeDist+2) {
    grabIndex = grabEnd + 2;
    grabDist = eDist;
  }
  
  else if (edgeDist+2 < cDist) {
    grabIndex = 5;
    grabDist = edgeDist;
  }

  else {
    grabIndex = grabCorner;
    grabDist = cDist;
  }

  switch (grabIndex)
    {
    case 1:
      grabPoint = frontEdge.start();
      rotAxis1 = convertTo3D(frontEdge.end(), 1);
      rotAxis2 = rotAxis1.reflect(midPoint_3d);
      break;

    case 2:
      grabPoint = frontEdge.end();
      rotAxis1 = convertTo3D(frontEdge.start(), 1);
      rotAxis2 = rotAxis1.reflect(midPoint_3d);
      break;
      
    case 3:
      grabPoint = midLine.start();
      rotAxis1 = convertTo3D(midLine.end(), 0);
      rotAxis2 = convertTo3D(frontEdge.end(), 1);
      protMode = true;              // might not want to rotate past the other end
      break;                        // **! PROTECTION MODE is not used in version 1.04

    case 4:
      grabPoint = midLine.end();
      rotAxis1 = convertTo3D(midLine.start(), 0);
      rotAxis2 = convertTo3D(frontEdge.start(), 1);
      protMode = true;             // might not want to rotate past the other end
      break;
      
    case 5:
      mouseSeg = frontEdge.perpendicular(mousePoint);
      grabPoint = mouseSeg.end();
      rotAxis1 = convertTo3D(midLine.start(), 0);
      rotAxis2 = convertTo3D(midLine.end(), 0);
      break;

    }

  leda_d3_segment axisSegment(rotAxis1, rotAxis2);
  
  rotateAxis = axisSegment;
  /*
  cerr << " Grabbed plane " << grabbedPlane << endl
       << " Front Edge    " << frontEdge << endl
       << " Midline       " << midLine << endl
       << " Mouse point   " << mousePoint << endl
       << " Grab point    " << grabPoint << endl
       << " Rotate axis   " << rotateAxis << endl
       << " Grab distance " << grabDist << endl
       << " Grab Index    " << grabIndex << endl;
  */
  return;

}




/***************************************************************
 *
 * pre: zPlane = { 0,1 }
 *
 *      if zPlane = 0, then project the 2D point onto (x,y,0)
 *
 *      if zPlane = 1, then project the 2D point onto (x,y,zVal)
 *        where zVal is a plane parallel to the viewport,
 *        located between the viewport and the user
 *
 *      zVal is global variable
 *   
 **************************************************************/
leda_d3_point convertTo3D(const leda_point & pointIn, int zPlane)
{
  
  // the point is in front of the viewport
  if (zPlane == 1) { 
    leda_d3_point pointOut(pointIn.xcoord(), pointIn.ycoord(), zVal);
    return pointOut;
  }

  // point is coincident to viewport
  else { 
    leda_d3_point pointOut(pointIn.xcoord(), pointIn.ycoord(), 0);
    return pointOut;
  }
  
}
  

  


/*******************************************************************
 *
 * This function is called whenever the HELP button
 *  is selected by the user.   This function calls
 *  recursive function 'showHelpPanel'
 *
 * PRE: takes a valid text filename, which is the
 *      help file to be displayed
 *      Assumes that HELP_FILE pathname is valid.
 *
 *
 * POST:  Checks the HELP_FILE path for existence
 *        of help file, and returns an error message if
 *        help file is not found.
 *
 *        Otherwise calls 'showHelpPanel' 
 *        When 'showHelpPanel returns to this function,
 *        this function returns to calling function. 
 *
 ******************************************************************/
void HelpFile(leda_string filename) {

   //   int MAX=(int)(screen_height / 42);
   bool displayPanel = true;
   
   // use the next 3 lines if you prefer to set the path variable externally
   /*   
   extern char * APAK_PATH;    
   leda_string tmp = (APAK_PATH != NULL) ? APAK_PATH : ".";
   */

   leda_string tmp("");
   if (tmp == "") tmp = ".";
   filename = tmp + "/Help/" + filename;
   ifstream inStream(filename);
   if (! inStream)    {               // if the file can't be opened
      leda_panel Ack("..::Error::..");
      Ack.text_item("Sorry, the specified help file was not found.");
      Ack.fbutton("&OK");
      while ( Ack.open() != 0 );
      return;
   }
   
   showHelpPanel(displayPanel, inStream);
   inStream.close();
   return; 
}


/****************************************************************
 *  Recursive function to open a text file onto a filestream,
 *  and page through the file, one panel at a time, until user
 *  chooses DONE.
 *
 *  Function calls itself recursively every time user presses
 *  'NEXT->' button, for next page of help text.
 *
 *  PRE: input arguments are:
 *       showPanel - boolean, must be true to open a new panel
 *                   function assigns false value when unwinding
 *
 *       maxLines - maximum number of text lines to display
 *
 *       inFileStream - name of file stream to which text file
 *                      is attached
 *
 * *****************************************************************/
void showHelpPanel(bool& showPanel, ifstream& inFileStream)
{
  char in_str[500];
  int maxLines = (int)(screen_height / 54);
  leda_panel help_Panel("Help");

  int lineCount = 0;
  
  while (showPanel) {
    while (! inFileStream.eof()) {
      if (lineCount > maxLines)
	break;
      lineCount++;
      inFileStream.getline(in_str, 2000);    
      help_Panel.text_item(in_str);
      help_Panel.text_item("");
    }
    
    help_Panel.fbutton("&DONE");  
    
    if (lineCount > maxLines)  {
      help_Panel.button("&NEXT->");
      int buttonPress = 99;
      while ((buttonPress = help_Panel.open()) != 0) {
	if (buttonPress == 1) {    // pressed NEXT, so open next page
	  lineCount = 0;
	  showHelpPanel(showPanel, inFileStream);
	}

	if (showPanel == false)
	  break;
      }
      showPanel = false;   // unwind the recursion
      help_Panel.close();
      return;              // return to calling function
    }

    else  {
      int buttonPress = 99;
      while ((buttonPress = help_Panel.open()) != 0)
	continue;
      showPanel = false;
    }
  }
  help_Panel.close();  // we get here for the special case
  return;              // if we only opened a single panel of text
}




/****************************************************************
 *
 *  Function displays information about ArrangePak
 *    includes a menu item for contact info
 *
 ****************************************************************/
 void displayAbout() {

   int z;
   leda_string panelName("ArrangePak-3D ver 1.04"); 

   leda_panel aboutPanel(700,210,panelName); 
   aboutPanel.text_item("");
   aboutPanel.text_item("");
   aboutPanel.text_item("");

   aboutPanel.text_item("ArrangePak-3D is designed for researchers studying  "
			"arrangements of planes in three-dimensional space. It provides "
			"interactive manipulation of planes, and export of the "
			"arrangement to a two-dimensional graph.");

   aboutPanel.text_item("");
   aboutPanel.text_item("");	 
   aboutPanel.text_item("A description of ArrangePak-3D is available at ");
   aboutPanel.text_item(" http://www.cs.uleth.ca/~wismath");
  
   aboutPanel.text_item("");
   aboutPanel.text_item("The software can be downloaded at");
   aboutPanel.text_item("http://www.cs.uleth.ca/~wismath/packages/ " );
   
   aboutPanel.text_item("");
   aboutPanel.text_item("This software is released on an \"AS IS\" basis, under ");
   aboutPanel.text_item("the terms of the GNU General Public Licence,");
   aboutPanel.text_item("except the Library of Efficient Data Types and Algorithms ");
   aboutPanel.text_item("which is licensed by Algorithmic Solutions Software GmbH. ");
   aboutPanel.text_item("");
   aboutPanel.button("Contact &Information");
   aboutPanel.fbutton("&Continue");
   
 while ((z = aboutPanel.open()) != 1)
   {
      if (z == 0) {        //  display contact info in text-box

	 leda_panel infoPan(500,320,panelName);

         infoPan.text_item("");
         infoPan.text_item("");
         infoPan.text_item("Designed by S.K. Wismath, Ph.D.");
         infoPan.text_item("");
         infoPan.text_item("--------------------------------------------------------------------------------");
         infoPan.text_item("");
         infoPan.text_item("Software development: Ray Dufresne");
         infoPan.text_item("");
         infoPan.text_item("Beta testing: Elspeth Nickle");
         infoPan.text_item("");
	 infoPan.text_item("");
         infoPan.text_item("");
         infoPan.text_item("Department of Mathematics and Computer Science");
         infoPan.text_item("");
         infoPan.text_item("University of Lethbridge");
         infoPan.text_item("");
         infoPan.text_item("4401 University Drive");
         infoPan.text_item("");
         infoPan.text_item("Lethbridge, AB  Canada  T1K 3M4");
         infoPan.text_item("");
         infoPan.text_item("e-mail: wismath@cs.uleth.ca");      
         infoPan.text_item("");
      	 infoPan.text_item("");
	 infoPan.text_item("");
         infoPan.text_item(" Compiled using LEDA, the Library of Efficient Data Types and Algorithms");
         infoPan.text_item("");
         infoPan.text_item("");
         infoPan.text_item("");
         infoPan.text_item("Copyright (c) 2004, 2003 Stephen Wismath");
	 infoPan.text_item("");
	 infoPan.text_item("--------------------------------------------------------------------------------");
         infoPan.text_item("");
         infoPan.text_item("This work was supported by N.S.E.R.C.");
         infoPan.fbutton("&Return");
         while (infoPan.open() != 0);
      }  
   }
}



/****************************************************************
 *
 *  Function displays a splash screen with message before 
 *  starting the application
 *
 ****************************************************************/
void displaySplashPanel()
{

   int z;
   leda_string panelName("ArrangePak-3D version 1.04"); 

   leda_panel aboutPanel(400,100,panelName); 
   aboutPanel.text_item("");
   aboutPanel.text_item("");
   aboutPanel.text_item("");

   aboutPanel.text_item("This is ArrangePak-3D version 1.04                ");
   aboutPanel.text_item("");
   aboutPanel.text_item("");
   aboutPanel.text_item("");
   aboutPanel.text_item("------------------------------------------------");
   aboutPanel.text_item("");
   aboutPanel.text_item("");
   aboutPanel.text_item("Bug reports to: wismath@cs.uleth.ca");
   aboutPanel.text_item("");
   aboutPanel.text_item("");
   aboutPanel.text_item("");
   aboutPanel.fbutton("&Continue");
   
   while ((z = aboutPanel.open()) != 0)
     ;

   return;
}
